/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
/*------------------------------------------------------------------------------
 *
 * OORT - load.c - Load data.
 *
 * $Id: load.c,v 1.2 1994/01/28 00:19:18 mtj Exp $
 * 
 * Chris Fouts - April, 1993.
 * 
 *----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <malloc.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <pf.h>
#include <prmath.h>

#include "oort.h"
#include "load.h"

#define	NEW_GROUP_TOKEN		1
#define	NEW_TMESH_TOKEN		2
#define	NEW_SMOOTH_TMESH_TOKEN	3
#define	NEW_GEO_STATE_TOKEN	4

#define	TERRAIN_GEO_STATE	0
#define	VEHICLE_GEO_STATE	1
#define	TIRES_GEO_STATE		2
#define	SHIELD_GEO_STATE	3
#define	N_GEO_STATES		4



extern pfVec3	groundColor ;
extern pfVec3	vehicleColor ;
extern char	*basename ;



/* BEGIN PROTOTYPES -S load.c */
static int              checkHeader( long fd, char *magic ) ;
static void             colorShield( void ) ;
static pfGeoState *     createShieldGeoState( void *arena ) ;
static pfGeoState *     createTerrainGeoState( void *arena ) ;
static pfGeoState *     createTiresGeoState( void *arena ) ;
static pfGeoState *     createVehicleGeoState( void *arena ) ;
static void             offsetCoords( float x, float y, float z, pfVec3 *c,
                            int n ) ;
static void             parseFile( long fd, pfGroup *head, pfVec3 *coords,
                            void *arena ) ;
static void             parseGroup( long fd, pfGroup *head, pfVec3 *coords,
                            void *arena ) ;
static void             parseTmesh( long fd, pfGeode *geode, pfVec3 *coords,
                            void *arena, int flat ) ;
static void             readAfterMalloc( long fd, void **data, int size ) ;
static int              readCoords( long fd, pfVec3 **c ) ;
static void             scaleCoords( float s, pfVec3 *c, int n ) ;
static ushort *         seqIndex( void ) ;
static void             setColorsFromNormals( pfGeoSet *gset, long nNorms,
                            pfVec3 *normals ) ;
static void             setupTexture( int fd, pfVec3 *v, int nCoords,
                            char *fileName ) ;
/* END PROTOTYPES -S load.c */

int			terrainTextured = 1 ;

static char		*dataFile ;
static pfVec3		*terrainCoords ;
static pfVec2		*texCoords ;
static int		nTCoords ;
static pfVec3		*vehicleCoords ;
static int		nVCoords ;
static pfVec3		*shieldCoords ;
static pfVec4		*shieldColor ;
static int		nSCoords ;
static pfGeoState	*geoState[N_GEO_STATES] ;
static pfGeoState	*currentGeoState ;
static pfVec4		*currentColor = NULL ;
static float		*currentCalcColor = NULL ;
static pfTexture	*currentTexture = NULL ;
static pfTexture	*terrainTex ;



/*------------------------------------------------------------------------------
 * Read in the terrain data.
 *----------------------------------------------------------------------------*/
pfGroup *
loadOortTerrain(
	char	*file
	)
{
	long	fd ;
	void	*arena ;
	pfGroup	*group ;
	pfDCS	*dcs ;

	dataFile = findFile( file ) ;

	currentColor = NULL ;
	currentCalcColor = groundColor ;

	currentGeoState = geoState[TERRAIN_GEO_STATE] ;

	/*
	 * Open the terrain file.
	 */
	if( ( fd = open( dataFile, O_RDONLY ) ) == -1 )
	{
		fprintf( stderr, "%s: Could not open `%s'\n", basename,
			dataFile ) ;
		free( dataFile ) ;
		return( NULL ) ;
	}

	/*
	 * Check the header.
	 */
	if( checkHeader( fd, "oort1.1" ) )
	{
		lseek( fd, 0, SEEK_SET ) ;
		if( checkHeader( fd, "oort1" ) )
		{
			fprintf( stderr, "%s: Incorrect header in `%s'\n",
				basename, dataFile );
			free( dataFile ) ;
			return( NULL ) ;
		}
		terrainTextured = 0 ;
	}

	/*
	 * Get the shared memory arena.
	 */
	arena = pfGetSharedArena() ;

	/*
	 * Read in the coordinates.
	 */
	if( ( nTCoords = readCoords( fd, &terrainCoords ) ) == 0 )
	{
		free( dataFile ) ;
		return( NULL ) ;
	}

	scaleCoords( ASTEROID_RADIUS, terrainCoords, nTCoords ) ;

	if( terrainTextured )
	{
		setupTexture( fd, terrainCoords, nTCoords, "terrain.bw" ) ;
	}

	/*
	 * Parse the rest of the file.
	 */
	group = pfNewGroup() ;

	parseFile( fd, group, terrainCoords, arena ) ;

	free( dataFile ) ;

	close( fd ) ;

	return( group ) ;
}



/*------------------------------------------------------------------------------
 * Create the geostate structure for the terrain.
 *----------------------------------------------------------------------------*/
static pfGeoState *
createTerrainGeoState(
	void	*arena
	)
{
	pfGeoState	*gs ;
	pfMaterial	*material ;

	/* 
	 * create a new geostate from shared memory,
	 * disable texturing and enable transparency
	 */
	gs = pfNewGState( arena ) ;
	pfGStateMode( gs, PFSTATE_ENTEXTURE, 1 ) ;
	pfGStateMode( gs, PFSTATE_TRANSPARENCY, 0 ) ;

	/* construct default geostate */
	pfGStateMode( gs, PFSTATE_ENLIGHTING, PF_OFF ) ;
	pfGStateMode( gs, PFSTATE_CULLFACE, PFCF_BACK ) ;

	return( gs ) ;
}



/*------------------------------------------------------------------------------
 * Check the header of a file versus a string.
 *----------------------------------------------------------------------------*/
static int
checkHeader(
	long	fd,
	char	*magic
	)
{
	int	n = strlen( magic ) + 1 ;
	char	*buffer ;

	if( ( buffer = malloc( n ) ) == NULL )
	{
		fprintf( stderr, "%s: out of memory.\n", basename ) ;
	}

	if( ( read( fd, buffer, n ) ) != n )
	{
		free( buffer ) ;
		return( 1 ) ;
	}

	n = strcmp( buffer, magic ) ;

	free( buffer ) ;

	return( n ) ;
}



/*------------------------------------------------------------------------------
 * Read in the coordinates.
 *----------------------------------------------------------------------------*/
static int
readCoords(
	long	fd,
	pfVec3	**c
	)
{
	int	nC ;

	readIn( fd, &nC, sizeof( nC ) ) ;
	if( nC < 3 )
	{
		return( 0 ) ;
	}

	/*
	 * Allocate space for the coordinates in the shared memory arena.
	 */
	*c = (pfVec3 *)myMalloc( nC * sizeof( pfVec3 ) ) ;

	readIn( fd, *c, nC * sizeof( pfVec3 ) ) ;

	return( nC ) ;
}



/*------------------------------------------------------------------------------
 * Read in from a file.
 *----------------------------------------------------------------------------*/
void
readIn(
	long	fd,
	void	*data,
	int	size
	)
{
	int	nRead ;

	if( ( nRead = read( fd, data, size ) ) != size )
	{
		fprintf( stderr, "%s: Error reading `%s' (read %d of %d "
			"bytes)\n", basename, dataFile, nRead, size ) ;
		pfExit() ;
		exit( -1 ) ;
	}
}



/*------------------------------------------------------------------------------
 * Parse the file for groups, tmeshes, and normals.
 *----------------------------------------------------------------------------*/
static void
parseFile(
	long		fd,
	pfGroup		*head,
	pfVec3		*coords,
	void		*arena
	)
{
	int	st ;
	char	token ;
	short	groupSize ;

	while( ( st = read( fd, &token, sizeof( token ) ) ) != 0 )
	{
		if( st != sizeof( token ) )
		{
			fprintf( stderr, "%s: Error reading `%s'\n",
				basename, dataFile ) ;
			pfExit() ;
			exit( -1 ) ;
		}

		if( token != NEW_GROUP_TOKEN )
		{
			fprintf( stderr, "%s: Unexpected data in `%s' (0x%02x),"
				" byte #0x%x\n", basename, dataFile,
				token, lseek( fd, 0, SEEK_CUR ) ) ;
			pfExit() ;
			exit( -1 ) ;
		}

		parseGroup( fd, head, coords, arena ) ;
	}
}



/*------------------------------------------------------------------------------
 * Parse a file to complete a group.
 *----------------------------------------------------------------------------*/
static void
parseGroup(
	long		fd,
	pfGroup		*head,
	pfVec3		*coords,
	void		*arena
	)
{
	int	i ;
	char	token ;
	short	groupSize ;
	pfGroup	*group ;
	pfGeode	*geode = NULL ;

	group = pfNewGroup() ;
	pfAddChild( head, group ) ;

	readIn( fd, &groupSize, sizeof( groupSize ) ) ;
	for( i = 0 ; i < groupSize ; i++ )
	{
		readIn( fd, &token, sizeof( token ) ) ;

		switch( token )
		{
			case NEW_GROUP_TOKEN :
				parseGroup( fd, group, coords, arena ) ;
				geode = NULL ;
				break ;

			case NEW_TMESH_TOKEN :
				if( geode == NULL )
				{
					geode = pfNewGeode() ;
					pfAddChild( group, geode ) ;
				}
				parseTmesh( fd, geode, coords, arena,
					PF_ON ) ;
				break ;

			case NEW_SMOOTH_TMESH_TOKEN :
				if( geode == NULL )
				{
					geode = pfNewGeode() ;
					pfAddChild( group, geode ) ;
				}
				parseTmesh( fd, geode, coords, arena,
					PF_OFF ) ;
				break ;

			case NEW_GEO_STATE_TOKEN :
				readIn( fd, &token, sizeof( token ) ) ;
				if( token >= 0 && token < N_GEO_STATES )
				{
					currentGeoState = geoState[token] ;
				}
				else
				{
					fprintf( stderr, "%s: bad "
						"geostate in `%s'\n",
						basename, dataFile ) ;
					endProgram( 1 ) ;
				}
				break ;

			default:
				fprintf( stderr, "%s: Unexpected data in "
					"`%s'\n", basename, dataFile ) ;
				endProgram( -1 ) ;
		}
	}
}



/*------------------------------------------------------------------------------
 * Parse a tmesh.
 *----------------------------------------------------------------------------*/
static void
parseTmesh(
	long		fd,
	pfGeode		*geode,
	pfVec3		*coords,
	void		*arena,
	int		flat
	)
{
	int		i ;
	short		nIndices ;
	ushort		*indices ;
	long		nStrips ;
	long		nNorms ;
	long		*primLens ;
	pfVec3		*normals ;
	pfGeoSet	*gset ;

	gset = pfNewGSet( arena ) ;

	readIn( fd, &nStrips, sizeof( nStrips ) ) ;

	if( nStrips < 1 )
	{
		fprintf( stderr, "%s: Unexpected data in `%s'\n", basename,
			dataFile ) ;
		pfExit() ;
		exit( -1 ) ;
	}
	readAfterMalloc( fd, (void **)&primLens, nStrips * sizeof( long ) ) ;

	nIndices = 0 ;
	for( i = 0 ; i < nStrips ; i++ )
	{
		nIndices += primLens[i] ;
	}

	readAfterMalloc( fd, (void **)&indices, nIndices * sizeof( ushort ) ) ;

	nNorms = nIndices ;

	readAfterMalloc( fd, (void **)&normals, nNorms * sizeof( pfVec3 ) ) ;

	pfGSetPrimType( gset, PFGS_TRISTRIPS ) ;
	pfGSetNumPrims( gset, nStrips ) ;
	pfGSetPrimLengths( gset, primLens ) ;
	pfGSetAttr( gset, PFGS_COORD3, PFGS_PER_VERTEX, coords,
			indices ) ;
	if( currentColor != NULL )
	{
		pfGSetAttr( gset, PFGS_COLOR4, PFGS_PER_VERTEX, currentColor,
				indices ) ;
	}
	else if( currentCalcColor != NULL )
	{
		setColorsFromNormals( gset, nNorms, normals ) ;
		pfFree( normals ) ;
	}
	else
	{
		pfGSetAttr( gset, PFGS_NORMAL3, PFGS_PER_VERTEX, normals,
				seqIndex() ) ;
	}

	if( currentTexture != NULL )
	{
		pfTexEnv	*tenv ;

		tenv = pfNewTEnv( arena ) ;
		pfTEnvMode( tenv, PFTE_MODULATE ) ;
		pfGSetAttr( gset, PFGS_TEXCOORD2, PFGS_PER_VERTEX, texCoords,
				indices ) ;
		pfGStateAttr( currentGeoState, PFSTATE_TEXTURE,
				currentTexture ) ;
		pfGStateAttr( currentGeoState, PFSTATE_TEXENV, tenv ) ;
	}

	pfGSetDrawMode( gset, PFGS_FLATSHADE, flat ) ;

	pfGSetGState( gset, currentGeoState ) ;

	pfAddGSet( geode, gset ) ;
}



/*------------------------------------------------------------------------------
 * Allocate space in the shared arena and read in data.
 *----------------------------------------------------------------------------*/
static void
readAfterMalloc(
	long	fd,
	void	**data,
	int	size
	)
{
	*data = myMalloc( size ) ;

	readIn( fd, *data, size ) ;
}



/*------------------------------------------------------------------------------
 * Scale the coordinates from unit values.
 *----------------------------------------------------------------------------*/
static void
scaleCoords(
	float	s,
	pfVec3	*c,
	int	n
	)
{
	int	i ;

	for( i = n - 1 ; i >= 0 ; i-- )
	{
		PFSCALE_VEC3( c[i], s, c[i] ) ;
	}
}



/*------------------------------------------------------------------------------
 * Offset the coordinates.
 *----------------------------------------------------------------------------*/
static void
offsetCoords(
	float	x,
	float	y,
	float	z,
	pfVec3	*c,
	int	n
	)
{
	int	i ;

	for( i = n - 1 ; i >= 0 ; i-- )
	{
		c[i][0] += x ;
		c[i][1] += y ;
		c[i][2] += z ;
	}
}



/*------------------------------------------------------------------------------
 * Read in the vehicle data.
 *----------------------------------------------------------------------------*/
pfGroup *
loadOortVehicle(
	char	*file
	)
{
	long	fd ;
	void	*arena ;
	pfGroup	*group ;

	dataFile = findFile( file ) ;

	currentColor = NULL ;
	currentCalcColor = NULL ;
	currentTexture = NULL ;

	/*
	 * Get the shared memory arena.
	 */
	arena = pfGetSharedArena() ;

	/*
	 * Create vehicle GeoState based on current vehicle color.
	 */
	geoState[VEHICLE_GEO_STATE] = createVehicleGeoState( arena ) ;

	/*
	 * Open the vehicle file.
	 */
	if( ( fd = open( dataFile, O_RDONLY ) ) == -1 )
	{
		fprintf( stderr, "%s: Could not open `%s'\n", basename,
			dataFile ) ;
		free( dataFile ) ;
		return( NULL ) ;
	}

	/*
	 * Check the header.
	 */
	if( checkHeader( fd, "oort-vehicle1" ) )
	{
		fprintf( stderr, "%s: Incorrect header in `%s'\n", basename,
			dataFile );
		free( dataFile ) ;
		return( NULL ) ;
	}

	/*
	 * Read in the coordinates.
	 */
	if( ( nVCoords = readCoords( fd, &vehicleCoords ) ) == 0 )
	{
		free( dataFile ) ;
		return( NULL ) ;
	}

	/*
	 * Parse the rest of the file.
	 */
	group = pfNewGroup() ;

	parseFile( fd, (pfGroup *)group, vehicleCoords, arena ) ;

	close( fd ) ;

	free( dataFile ) ;

	return( group ) ;
}



/*------------------------------------------------------------------------------
 * Create the geostate structure for the vehicle body.
 *----------------------------------------------------------------------------*/
static pfGeoState *
createVehicleGeoState(
	void	*arena
	)
{
	pfGeoState	*gs ;
	pfMaterial	*material ;

	/* 
	 * create a new geostate from shared memory,
	 * disable texturing and enable transparency
	 */
	gs = pfNewGState( arena ) ;
	pfGStateMode( gs, PFSTATE_ENTEXTURE, 0 ) ;
	pfGStateMode( gs, PFSTATE_TRANSPARENCY, 0 ) ;

	/* construct default material */
	material = pfNewMtl( arena ) ;
	pfMtlColor( material, PFMTL_AMBIENT, 0.1f*vehicleColor[0],
			0.1f*vehicleColor[1], 0.1f*vehicleColor[2] ) ;
	pfMtlColor( material, PFMTL_DIFFUSE, vehicleColor[0], vehicleColor[1],
			vehicleColor[2] ) ;
	pfMtlColorMode( material, PFMTL_FRONT, LMC_COLOR ) ;

	/* construct default geostate */
	pfGStateAttr( gs, PFSTATE_FRONTMTL, material ) ;
	pfGStateMode( gs, PFSTATE_ENLIGHTING, PF_ON ) ;
	pfGStateMode( gs, PFSTATE_CULLFACE, PFCF_BACK ) ;

	return( gs ) ;
}



/*------------------------------------------------------------------------------
 * Create the geostate structure for the vehicle tires.
 *----------------------------------------------------------------------------*/
static pfGeoState *
createTiresGeoState(
	void	*arena
	)
{
	pfGeoState	*gs ;
	pfMaterial	*material ;

	/* 
	 * create a new geostate from shared memory,
	 * disable texturing and enable transparency
	 */
	gs = pfNewGState( arena ) ;
	pfGStateMode( gs, PFSTATE_ENTEXTURE, 0 ) ;
	pfGStateMode( gs, PFSTATE_TRANSPARENCY, 0 ) ;

	/* construct default material */
	material = pfNewMtl( arena ) ;
	pfMtlColor( material, PFMTL_AMBIENT, 0.02f, 0.02f, 0.02f ) ;
	pfMtlColor( material, PFMTL_DIFFUSE, 0.10f, 0.10f, 0.10f ) ;
	pfMtlColorMode( material, PFMTL_FRONT, LMC_COLOR ) ;

	/* construct default geostate */
	pfGStateAttr( gs, PFSTATE_FRONTMTL, material ) ;
	pfGStateMode( gs, PFSTATE_ENLIGHTING, PF_ON ) ;
	pfGStateMode( gs, PFSTATE_CULLFACE, PFCF_BACK ) ;

	return( gs ) ;
}



/*------------------------------------------------------------------------------
 * Return (after generating if necessary) a sequential index that goes from
 * 0 to the maximum unsigned short value.
 *----------------------------------------------------------------------------*/
static ushort *
seqIndex(
	void
	)
{
	int		i ;
	int		n ;
	static ushort	*sI = NULL ;

	if( sI == NULL )
	{
		n = ( 1 << ( sizeof( ushort ) * 8 ) ) - 1 ;
		sI = (ushort *)myMalloc( n * sizeof( ushort ) ) ;
		for( i = 0 ; i < n ; i++ )
		{
			sI[i] = i ;
		}
	}

	return( sI ) ;
}



/*------------------------------------------------------------------------------
 * Create the geo states.
 *----------------------------------------------------------------------------*/
void
initGeoStates(
	void
	)
{
	void	*arena ;

	arena = pfGetSharedArena();

	geoState[TERRAIN_GEO_STATE] = createTerrainGeoState( arena ) ;
	geoState[TIRES_GEO_STATE] = createTiresGeoState( arena ) ;
	geoState[SHIELD_GEO_STATE] = createShieldGeoState( arena ) ;

	currentGeoState = geoState[TERRAIN_GEO_STATE] ;
}



/*------------------------------------------------------------------------------
 * Read in the shield data.
 *----------------------------------------------------------------------------*/
pfGroup *
loadOortShield(
	char	*file
	)
{
	long	fd ;
	void	*arena ;
	pfGroup	*group ;

	dataFile = findFile( file ) ;

	currentGeoState = geoState[SHIELD_GEO_STATE] ;

	/*
	 * Open the shield file.
	 */
	if( ( fd = open( dataFile, O_RDONLY ) ) == -1 )
	{
		fprintf( stderr, "%s: Could not open `%s'\n", basename,
			dataFile ) ;
		free( dataFile ) ;
		return( NULL ) ;
	}

	/*
	 * Check the header.
	 */
	if( checkHeader( fd, "shield1" ) )
	{
		fprintf( stderr, "%s: Incorrect header in `%s'\n", basename,
			dataFile ) ;
		free( dataFile ) ;
		return( NULL ) ;
	}

	/*
	 * Get the shared memory arena.
	 */
	arena = pfGetSharedArena() ;

	/*
	 * Read in the coordinates.
	 */
	if( ( nSCoords = readCoords( fd, &shieldCoords ) ) == 0 )
	{
		free( dataFile ) ;
		return( NULL ) ;
	}

	/*
	 * Create colors for shields.
	 */
	colorShield() ;

	currentColor = shieldColor ;
	currentCalcColor = NULL ;
	currentTexture = NULL ;

	/*
	 * Parse the rest of the file.
	 */
	group = pfNewGroup() ;

	parseFile( fd, group, shieldCoords, arena ) ;

	scaleCoords( SHIELD_RADIUS, shieldCoords, nSCoords ) ;
	offsetCoords( 0.0f, -CAR_LENGTH / 6.0f, 0.5f * CAR_HEIGHT,
			shieldCoords, nSCoords ) ;

	free( dataFile ) ;

	close( fd ) ;

	return( group ) ;
}



/*------------------------------------------------------------------------------
 * Create the geostate structure for the shield body.
 *----------------------------------------------------------------------------*/
static pfGeoState *
createShieldGeoState(
	void	*arena
	)
{
	pfGeoState	*gs ;

	/* 
	 * create a new geostate from shared memory,
	 * disable texturing and enable transparency
	 */
	gs = pfNewGState( arena ) ;
	pfGStateMode( gs, PFSTATE_ENTEXTURE, 0 ) ;
	pfGStateMode( gs, PFSTATE_TRANSPARENCY, PFTR_ON ) ;

	/* construct default geostate */
	pfGStateMode( gs, PFSTATE_ENLIGHTING, PF_OFF ) ;
	pfGStateMode( gs, PFSTATE_CULLFACE, PFCF_BACK ) ;

	return( gs ) ;
}



/*------------------------------------------------------------------------------
 * Add color to shields.
 *----------------------------------------------------------------------------*/
static void
colorShield(
	void
	)
{
	int	i ;
	float	v ;
	pfVec4	*c ;

	shieldColor = (pfVec4 *)myMalloc( nSCoords * sizeof( pfVec4 ) ) ;

	for( i = 0 ; i < nSCoords ; i++ )
	{
		v = ABS( shieldCoords[i][2] ) ;
		PFSET_VEC4( shieldColor[i], 0.0f, 1.0f, 0.25f + v*0.75f, 0.25f );
	}
}




/*------------------------------------------------------------------------------
 * Calculate and set the colors based on the normals for a geoset.
 *----------------------------------------------------------------------------*/
static void
setColorsFromNormals(
	pfGeoSet	*gset,
	long		nNorms,
	pfVec3		*normals
	)
{
	int		i ;
	float		f ;
	pfVec4		*colors ;

	colors = (pfVec4 *)myMalloc( nNorms * sizeof( pfVec4 ) ) ;

	for( i = 0 ; i < nNorms ; i++ )
	{
		f = -normals[i][1] ;
		if( f < 0.0f )
		{
			PFSET_VEC4( colors[i], 0.0f, 0.0f, 0.0f, 1.0f ) ;
		}
		else
		{
			PFSET_VEC4( colors[i],
					f * currentCalcColor[0],
					f * currentCalcColor[1],
					f * currentCalcColor[2],
					1.0f ) ;
		}
	}

	pfGSetAttr( gset, PFGS_COLOR4, PFGS_PER_VERTEX, colors, seqIndex() ) ;
}



/*------------------------------------------------------------------------------
 * Calculate the texture coords for the asteroid.
 *----------------------------------------------------------------------------*/
static void
setupTexture(
	int	fd,
	pfVec3	*v,
	int	nCoords,
	char	*fileName
	)
{
	int	i ;
	float	f = 32.0f ;
	char	*texFile ;

	texFile = findFile( fileName ) ;

	/*
	 * Read the texture coordinate data file.
	 */
	readAfterMalloc( fd, (void **)&texCoords, nCoords * sizeof( pfVec2 ) ) ;

	for( i = 0 ; i < nCoords ; i++ )
	{
		texCoords[i][0] = fmodf( f * texCoords[i][0], 1.0f ) ;
		texCoords[i][1] = fmodf( f * texCoords[i][1], 1.0f ) ;
	}

	terrainTex = pfNewTex( pfGetSharedArena() ) ;

	if( pfLoadTexFile( terrainTex, texFile ) == 0 )
	{
		currentTexture = NULL ;
		pfDelete( terrainTex ) ;
		terrainTex = NULL ;
	}
	else
	{
		currentTexture = terrainTex ;
	}

	free( texFile ) ;
}



