/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
/*------------------------------------------------------------------------------
 *
 * OORT - sound.c - Routines for sounds.
 *
 * $Id: sound.c,v 1.2 1994/01/28 00:21:19 mtj Exp $
 * 
 * Chris Fouts - May, 1993.
 * 
 *----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <errno.h>
#include <string.h>
#include <malloc.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/prctl.h>
#include <invent.h>
#include <signal.h>

#include "audio.h"
#include "audiofile.h"

#include "oort.h"
#include "sound.h"

#define	MAX_AUDIO_PORTS	3

static ALport		audioPort[MAX_AUDIO_PORTS] ;
static int		nAudioPorts ;
static char		*audioDir ;
static int		soundEnabled = 0 ;
static long		soundSize[SFX_N_SOUNDS] ;
static short		*soundData[SFX_N_SOUNDS] ;
static int		soundChild ;
static short		*newSoundData ;
static long		newSoundSize ;
static int		newSound = 0 ;
static unsigned long	origLeftGain ;
static unsigned long	origRightGain ;
static unsigned long	origOutputRate ;
static unsigned long	currLeftGain ;
static unsigned long	currRightGain ;
static unsigned long	gainSettings[] =
				{ 0, 2, 3, 5, 9, 16, 29, 50, 91, 156, 255 } ;
static int		gainIndex ;
static struct sigaction	parentAct ;
static int		endingOnPurpose = 0 ;
static char		*soundFile[] = {
				"fire.sound",
				"hit.sound",
				"ok.sound",
				"not_ok.sound",
				"static.sound",
				"klaxon.sound",
				"newmail.sound",
				} ;


extern char	*basename ;

/* BEGIN PROTOTYPES -S sound.c */
static int      checkForAudio( void ) ;
static void     dieGracefully( int sig, int code, struct sigcontext *sc ) ;
static ALport   initAudioPort( int init ) ;
static short *  loadAudioFile( char *dirName, char *fileName, long *size ) ;
static int      openAudioPorts( ALport ports[], int maxAudioPorts ) ;
static void     resetAudioHw( void ) ;
static void     setGain( float l, float r ) ;
static void     setGainIndex( unsigned long gain ) ;
static void     signalSound( short *data, long size ) ;
static void     soundDied( int sig, int code, struct sigcontext *sc ) ;
static void     soundErrFunc( long err, const char *msg, ... ) ;
static void     soundHandler( void *arg ) ;
static void     wakeUp( int sig, int code, struct sigcontext *sc ) ;
/* END PROTOTYPES -S sound.c */



/*------------------------------------------------------------------------------
 * Initialize audio.
 *----------------------------------------------------------------------------*/
int
initSound(
	void
	)
{
	int	i ;
	char	*audio_dir ;

	(void)ALseterrorhandler( soundErrFunc ) ;

	nAudioPorts = openAudioPorts( audioPort, MAX_AUDIO_PORTS ) ;

	if( nAudioPorts == 0 )
	{
		fprintf( stderr, "%s: no audio ports available\n",
			basename ) ;
		return( 1 ) ;
	}
	else if( nAudioPorts < 0 )
	{
		return( 1 ) ;
	}


	if( ( audioDir = getenv( "OORT_DIR" ) ) == NULL )
	{
		audioDir = DATA_DIR ;
	}

	for( i = 0 ; i < SFX_N_SOUNDS ; i++ )
	{
		soundData[i] = loadAudioFile( audioDir, soundFile[i],
						&soundSize[i] ) ;
	}

	parentAct.sa_handler = soundDied ;
	sigemptyset( &(parentAct.sa_mask) ) ;
	parentAct.sa_flags = 0 ;
	sigaction( SIGCLD, &parentAct, NULL ) ;

	if( ( soundChild = sproc( soundHandler, PR_SALL, audioPort ) ) < 0 )
	{
		fprintf( stderr, "%s: could not sproc sound handler: "
				"%s\n", basename, strerror( errno ) ) ;
		return( 1 ) ;
	}
	else
	{
		soundEnabled = 1 ;
	}

	return( 0 ) ;
}



/*------------------------------------------------------------------------------
 * Open up the audio ports.
 *----------------------------------------------------------------------------*/
static int
openAudioPorts(
	ALport	ports[],
	int	maxAudioPorts
	)
{
	int	n ;

	if( checkForAudio() != 0 )
	{
		return( -1 ) ;
	}

	for( n = 0 ; n < maxAudioPorts ; n++ )
	{
		if( ( audioPort[n] = initAudioPort( n == 0 ) ) == NULL )
		{
			break ;
		}
	}

	return( n ) ;
}



/*------------------------------------------------------------------------------
 * Sound handler routine.
 *----------------------------------------------------------------------------*/
static void
soundHandler(
	void	*arg
	)
{
	ALport			*ap = (ALport *)arg ;
	int			i ;
	int			empty ;
	int			nextPort = 0 ;
	long			maxSampsPerFrame ;
	long			sampsToPlay[MAX_AUDIO_PORTS] ;
	long			sampsPlayed[MAX_AUDIO_PORTS] ;
	short			*sample[MAX_AUDIO_PORTS] ;
	struct sigaction	act ;

	prctl( PR_TERMCHILD, 0 ) ;

	act.sa_handler = wakeUp ;
	sigemptyset( &(act.sa_mask) ) ;
	act.sa_flags = 0 ;
	sigaction( SIGUSR1, &act, NULL ) ;

	act.sa_handler = dieGracefully ;
	sigemptyset( &(act.sa_mask) ) ;
	act.sa_flags = 0 ;
	sigaction( SIGHUP, &act, NULL ) ;

	maxSampsPerFrame = 1600 ;

	for( i = 0 ; i < nAudioPorts ; i++ )
	{
		sample[i] = 0 ;
		sampsToPlay[i] = 0 ;
	}

	while( 1 )
	{
		if( newSound )
		{
			sample[nextPort] = newSoundData ;
			sampsToPlay[nextPort] = newSoundSize ;
			sampsPlayed[nextPort] = 0 ;
			newSound = 0 ;
			nextPort = ( nextPort + 1 ) % nAudioPorts ;
		}
		empty = 1 ;
		for( i = 0 ; i < nAudioPorts ; i++ )
		{
			if( sampsToPlay[i] > maxSampsPerFrame )
			{
				setGain( 1., 1. ) ;
				ALwritesamps( ap[i],
						sample[i] + sampsPlayed[i],
						maxSampsPerFrame ) ;
				sampsPlayed[i] += maxSampsPerFrame ;
				sampsToPlay[i] -= maxSampsPerFrame ;
				empty = 0 ;
			}
			else if( sampsToPlay[i] > 0 )
			{
				setGain( 1., 1. ) ;
				ALwritesamps( ap[i], sample[i] + sampsPlayed[i],
						sampsToPlay[i] ) ;
				sampsToPlay[i] = 0 ;
			}
		}
		if( empty )
		{
			pause() ;
		}
	}
}



/*------------------------------------------------------------------------------
 * Empty error handler function for SIGUSR1.
 *----------------------------------------------------------------------------*/
static void
wakeUp(
	int			sig,
	int			code,
	struct sigcontext	*sc
	)
{
}



/*------------------------------------------------------------------------------
 * Empty error handler function for SIGUSR1.
 *----------------------------------------------------------------------------*/
static void
dieGracefully(
	int			sig,
	int			code,
	struct sigcontext	*sc
	)
{
	printf( "dying gracefully\n" ) ;
	exit( 0 ) ;
}



/*------------------------------------------------------------------------------
 * Clean up sound routines.
 *----------------------------------------------------------------------------*/
void
endSound(
	void
	)
{
	endingOnPurpose = 1 ;

	if( soundChild > 0 )
	{
		kill( soundChild, SIGKILL ) ;
	}
	if( nAudioPorts > 0 )
		resetAudioHw() ;
}



/*------------------------------------------------------------------------------
 * Reset the audio hardware to where we found it when we started.
 *----------------------------------------------------------------------------*/
static void
resetAudioHw(
	void
	)
{
	long	pvbuf[6] ;
	long	buflen ;

	pvbuf[0] = AL_LEFT_SPEAKER_GAIN ;
	pvbuf[1] = origLeftGain ;
	pvbuf[2] = AL_RIGHT_SPEAKER_GAIN ;
	pvbuf[3] = origRightGain ;
	pvbuf[4] = AL_OUTPUT_RATE ;
	pvbuf[5] = origOutputRate ;
	buflen = 6 ;

	ALsetparams( AL_DEFAULT_DEVICE, pvbuf, buflen ) ;
}



/*------------------------------------------------------------------------------
 * Open and initialize an audio port.
 *----------------------------------------------------------------------------*/
static ALport
initAudioPort(
	int	init
	)
{
	long		pvbuf[6] ;
	long		buflen ;
	ALconfig	audioPortConfig;
	ALport		ap ;

	if( init )
	{
		pvbuf[0] = AL_LEFT_SPEAKER_GAIN ;
		pvbuf[2] = AL_RIGHT_SPEAKER_GAIN ;
		pvbuf[4] = AL_OUTPUT_RATE ;
		buflen = 6 ;
		ALgetparams( AL_DEFAULT_DEVICE, pvbuf, buflen ) ;
		currLeftGain = origLeftGain = pvbuf[1] ;
		currRightGain = origRightGain = pvbuf[3] ;
		origOutputRate = pvbuf[5] ;

		setGainIndex( ( origLeftGain + origRightGain ) / 2 ) ;
	}

	/*
	 * configure and open audio port
	 */
	audioPortConfig = ALnewconfig();
	ALsetwidth( audioPortConfig, AL_SAMPLE_16 ) ;
	ALsetchannels( audioPortConfig, AL_STEREO ) ;

	/*
	 * make the ring buffer large enough to hold 0.1 sec of audio samples
	 */
	ALsetqueuesize( audioPortConfig, 1600 ) ;
	ap = ALopenport( basename, "w", audioPortConfig ) ;

	ALfreeconfig( audioPortConfig ) ;

	return( ap ) ;
}



/*------------------------------------------------------------------------------
 * Set the gain index.
 *----------------------------------------------------------------------------*/
static void
setGainIndex(
	unsigned long	gain
	)
{
	int	i ;
	long	min = 256 ;
	long	diff ;

	for( i = 0 ; i < sizeof( gainSettings ) / sizeof( gainSettings[0] ) ;
		i++ )
	{
		diff = gain - gainSettings[i] ;
		if( diff < 0 )
			diff = -diff ;
		if( diff < min )
		{
			min = diff ;
			gainIndex = i ;
		}
	}
}



/*------------------------------------------------------------------------------
 * Set the volume level.
 *----------------------------------------------------------------------------*/
static void
setGain(
	float l,
	float r
	)
{
	long	pvbuf[6] ;
	long	buflen ;

	pvbuf[0] = AL_LEFT_SPEAKER_GAIN ;
	pvbuf[1] = l * currLeftGain ;
	pvbuf[2] = AL_RIGHT_SPEAKER_GAIN ;
	pvbuf[3] = r * currRightGain ;
	pvbuf[4] = AL_OUTPUT_RATE ;
	pvbuf[5] = AL_RATE_16000 ;
	buflen = 6 ;

	ALsetparams( AL_DEFAULT_DEVICE, pvbuf, buflen ) ;
}



/*------------------------------------------------------------------------------
 * Check for audio hardware.
 *----------------------------------------------------------------------------*/
static int
checkForAudio(
	void
	)
{
	int		st = 0 ;
	inventory_t	*base ;
	inventory_t	*inv ;

	inv = base = getinvent() ;
	while( inv != NULL )
	{
		/*
		 * Ok if any type of audio hardware available.
		 */
#if defined( __SVR3 )
		if( inv->class == INV_AUDIO )
#else
		if( inv->inv_class == INV_AUDIO )
#endif /* defined( __SVR3 ) */
		{
			st = 1 ;
		}
		inv = getinvent() ;
	}
	
	if( base )
		endinvent() ;

	return( st ) ;
}



/*------------------------------------------------------------------------------
 * Open an audio file, check for recognized type, and read.
 *----------------------------------------------------------------------------*/
static short *
loadAudioFile(
	char	*dirName,
	char	*fileName,
	long	*size
	)
{
	int		fd ;
	AFfilehandle	file ;
	int		framesPerSec ;
	long		bitsPerSamp ;
	long		sampType ;
	char		*name ;
	short		*data ;

	name = malloc( strlen( dirName ) + strlen( fileName ) + 2 ) ;
	sprintf( name, "%s/%s", dirName, fileName ) ;

	fd = open( name, O_RDONLY ) ;

	if( fd < 0 )
	{
		perror( name ) ;
		free( name ) ;
		return( NULL ) ;
	}

	switch( AFidentifyfd( fd ) )
	{
		case AF_FILE_AIFF :
		case AF_FILE_AIFFC :
			break ;

		default :
			fprintf( stderr, "%s: unrecognized file type -- convert"
				" to AIFF or AIFC\n", name ) ;
			free( name ) ;
			return( NULL ) ;
	}

	file = AFopenfd( fd, "r", AF_NULL_FILESETUP ) ;

	if( file == AF_NULL_FILEHANDLE )
	{
		fprintf( stderr, "%s: failed to attach an audio file struct\n",
			 name ) ;
		free( name ) ;
		return( NULL ) ;
	}

	if( (int)AFgetchannels( file, AF_DEFAULT_TRACK ) != 2 )
	{
		fprintf( stderr, "%s: does not have 2 channels\n", name ) ;
		free( name ) ;
		AFclosefile( file ) ;
		return( NULL ) ;
		}

	if( ( framesPerSec = AFgetrate( file, AF_DEFAULT_TRACK ) ) != 16000 )
	{
		fprintf( stderr, "%s: is not recorded at 16 KHz\n", name ) ;
		free( name ) ;
		AFclosefile( file ) ;
		return( NULL ) ;
	}

	AFgetsampfmt( file, AF_DEFAULT_TRACK, &sampType, &bitsPerSamp ) ;
	if( bitsPerSamp != 16 )
	{
		fprintf( stderr, "%s: is not recorded with 16 bit samples\n",
			name ) ;
		free( name ) ;
		AFclosefile( file ) ;
		return( NULL ) ;
	}

	free( name ) ;

	*size = AFgetframecnt( file, AF_DEFAULT_TRACK ) ;

	if( ( data = (short *)malloc( *size * 2 * sizeof( short ) ) ) == NULL )
	{
		fprintf( stderr, "%s: out of memory for %d samples\n",
			name, *size ) ;
		*size = 0 ;
		AFclosefile( file ) ;
		return( NULL ) ;
	}

	if( AFreadframes( file, AF_DEFAULT_TRACK, data, *size ) != *size )
	{
		fprintf( stderr, "%s: error reading\n", name ) ;
		free( data ) ;
		data = NULL ;
		*size = 0 ;
		}

	*size *= 2 ;

	AFclosefile( file ) ;

	return( data ) ;
}



/*------------------------------------------------------------------------------
 * Play sound effect.
 *----------------------------------------------------------------------------*/
void
sfx(
	int	sound
	)
{
	long	pvbuf[6] ;
	long	buflen ;

	/*
	 * If audio is possible.
	 */
	if( nAudioPorts > 0 && soundData[sound] && soundEnabled )
	{
		/*
		 * Check to see if the volume was changed externally.
		 */
		pvbuf[0] = AL_LEFT_SPEAKER_GAIN ;
		pvbuf[2] = AL_RIGHT_SPEAKER_GAIN ;
		buflen = 4 ;

		ALgetparams( AL_DEFAULT_DEVICE, pvbuf, buflen ) ;
		if( pvbuf[1] != currLeftGain || pvbuf[3] != currRightGain )
		{
			currLeftGain = pvbuf[1] ;
			currRightGain = pvbuf[3] ;
			setGainIndex( ( pvbuf[1] + pvbuf[3] ) / 2 ) ;
		}

		signalSound( soundData[sound], soundSize[sound] ) ;
	}
}



/*------------------------------------------------------------------------------
 * Wake up the sound handler.
 *----------------------------------------------------------------------------*/
static void
signalSound(
	short	*data,
	long	size
	)
{
	if( soundChild )
	{
		newSoundData = data ;
		newSoundSize = size ;
		newSound = 1 ;
		kill( soundChild, SIGUSR1 ) ;
	}
}



/*------------------------------------------------------------------------------
 * Adjust the volume.
 *----------------------------------------------------------------------------*/
int
volumeChange(
	int	direction
	)
{
	if( soundEnabled )
	{
		if( direction == 1 )
		{
			if( gainIndex < sizeof( gainSettings ) /
				sizeof( gainSettings[0] ) - 1 )
			{
				gainIndex++ ;
			}
		}
		else
		{
			if( gainIndex > 0 )
			{
				gainIndex-- ;
			}
		}
		currLeftGain = currRightGain = gainSettings[gainIndex] ;

		setGain( 1., 1. ) ;

		return( gainIndex ) ;
	}
	else
	{
		return( 0 ) ;
	}
}



/*------------------------------------------------------------------------------
 * Handler for sound child dying.
 *----------------------------------------------------------------------------*/
static void
soundDied(
	int			sig,
	int			code,
	struct sigcontext	*sc
	)
{
	if( !endingOnPurpose )
	{
		nAudioPorts = 0 ;
		soundEnabled = 0 ;

		fprintf( stderr, "Sound handling child process was killed\n" ) ;
	}
}



/*------------------------------------------------------------------------------
 * Set the volume to a specified level.
 *----------------------------------------------------------------------------*/
void
setVolume(
	int	level
	)
{
	if( level >= 0 &&
		level < sizeof( gainSettings ) / sizeof( gainSettings[0] ) )
	{
		gainIndex = level ;
		currLeftGain = currRightGain = gainSettings[gainIndex] ;
	}
}



/*------------------------------------------------------------------------------
 * Audio error function handler.
 *----------------------------------------------------------------------------*/
static void
soundErrFunc(
	long		err,
	const char	*msg,
	...
	)
{
}
