/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
/* this file takes care of all the intersection stuff between objects */
#include <stdio.h>
#include <gl.h>
#include "defs.h"
#include "camera.h"
#include "field.h"
#include "goals.h"
#include "creatures.h"
#include "game.h"

extern CREATURE creatures[MAX_PLAYERS];
extern VECTOR looking;   /* direction creature is looking */
extern CHIPS *chips[MAX_PLAYERS][CHIPS_PER_PLAYER];  /* the playing chips */
extern TRIANGLE hole_edge[12], hole_center[12];  /* triangulated hole */ 
extern VECTOR edge_border[6], center_border[6];  /* hole borders */
extern VECTOR bound[4][2];      /* how far the creature can go */
extern TRI_LIST *first_triangle[MAX_PLAYERS*2]; /* triangles of goals */ 
extern VECTOR goal_start[3][2];
extern VECTOR goal_bounds[3][2][3];  /* about where each goal sits */
extern TRI_LIST *first_triangle_intersect[MAX_PLAYERS*2]; /* intersections with
							     goals */
extern short me;  /* index of my creature */
extern short num_players; /* how many creatures are there */

/*
 * This function expects three points and returns the direction (clockwise or
 * counterclockwise ) traced out when moving from p0 to p1 to p2. If p0,p1,p2
 * lie on the same line then it can return one of three things depending on
 * their relative position.  If p2 is between p0 and p1 then they are
 * considered collinear. If p0 is between p1 and p2 then consider it clockwise
 * otherwise counterclockwise. 
 * (borrowed from dartmouth's ncmachining project)
 */

int
turn(VECTOR p0, VECTOR p1, VECTOR p2)
     
{
  double          dx1, dx2, dy1, dy2;
  
  dx1 = p1[0] - p0[0];
  dy1 = p1[2] - p0[2];
  dx2 = p2[0] - p0[0];
  dy2 = p2[2] - p0[2];
  
  if (dx1 * dy2 > dy1 * dx2)  /* slope of p0p2 is greater than slope of
			       * p0p1 */
    return (COUNTERCLOCK);
  else if (dx1 * dy2 < dy1 * dx2)
    return (CLOCKWISE);
  else
    {
      if ((dx1 * dx2 < 0) || (dy1 * dy2 < 0))
        return (CLOCKWISE);
      else if ((dx1 * dx1 + dy1 * dy1) >= (dx2 * dx2 + dy2 * dy2))
        return (COLLINEAR);
      else
        return (COUNTERCLOCK);
    }
}

/* do the bounding boxes defined by these points intersect?  */
int
bounding_box(VECTOR p1, VECTOR p2, VECTOR p3, VECTOR p4)
{
  float x1, x2, x3, x4;
  float y1, y2, y3, y4;
  
  x1 = MIN(p1[0], p2[0]);
  x2 = MAX(p1[0], p2[0]);
  x3 = MIN(p3[0], p4[0]);
  x4 = MAX(p3[0], p4[0]);

  y1 = MIN(p1[2], p2[2]);
  y2 = MAX(p1[2], p2[2]);
  y3 = MIN(p3[2], p4[2]);
  y4 = MAX(p3[2], p4[2]);
  
  if ((x2 >= x3) && (x4 >= x1) && (y2 >= y3) && (y4 >= y1))
    return(TRUE); /* bounding boxes intersect */

  else return(FALSE);
}


/* do the two line segments intersect? */
int 
seg_intersect(x1, x2, y1, y2)
     VECTOR x1, x2, y1, y2;
{
  return((turn(x1, x2, y1) != turn(x1, x2, y2)) &&
	 (turn(y1, y2, x1) != turn(y1, y2, x2)));
}


/* is the point in the triangle? */
int
in_triangle(VECTOR p, VECTOR t1, VECTOR t2, VECTOR t3)
{
  int dir1, dir2, dir3;
  
  dir1 = turn(t1, t2, p);
  dir2 = turn(t2, t3, p);
  dir3 = turn(t3, t1, p);
  
  return(((dir1 == dir2) && (dir2 == dir3)) || (dir1 == COLLINEAR) ||
	 (dir2 == COLLINEAR) || (dir3 == COLLINEAR));
}

/* does the triangle intersect the line? */
int
intersect_tl(VECTOR t1, VECTOR t2, VECTOR t3, 
	     VECTOR l1, VECTOR l2)
{
  int i;
    
  return(  (bounding_box(l1, l2, t1, t2) && seg_intersect(l1, l2, t1, t2))
	 || (bounding_box(l1, l2, t2, t3) && seg_intersect(l1, l2, t2, t3))
	 || (bounding_box(l1, l2, t3, t1) && seg_intersect(l1, l2, t3, t1))
	 || in_triangle(l1, t1, t2, t3)
	 || in_triangle(l2, t1, t2, t3));
  
}

/* is the point within the six-sided polygon? */
int 
in_polygon(VECTOR point, VECTOR polygon[6]) 
{
  int i, turn_result;
  VECTOR center = {0, 0, 0};
  int direction;
  
  direction = turn(polygon[0], polygon[1], point);

  for (i=0; i<5; i++)
  {
    if ((turn(point, center, polygon[i]) == CLOCKWISE) &&
	(turn(point, center, polygon[i+1]) == COUNTERCLOCK))
    {
      if (turn(polygon[i], polygon[i+1], point) == CLOCKWISE)
	return(TRUE); /* is within */
      else return(FALSE);
    }
  }
  
  /* it must be in the last segment */
  if ((turn(point, center, polygon[5]) == CLOCKWISE) &&
      (turn(point, center, polygon[0]) == COUNTERCLOCK))
    if(turn(polygon[5], polygon[0], point) == CLOCKWISE)
      return(TRUE);
  
  return(FALSE);
}


/* is the creatures' head banging against the wall? */
int
  border_test(int *bound1, int *bound2, int cindex)
{
  int i=0;
  VECTOR v;
  
  VEC_COPY(v, looking);
  VEC_NORM(v);
  VEC_MULT(v, -15);
  VEC_ADD(v, v, creatures[cindex].position);
  
  if (creatures[cindex].position[0] == 80)
  {
    if(seg_intersect(creatures[cindex].position, v, bound[0][0], bound[3][0]))
    {
      *bound1 = 3; *bound2 = 0;
      return(TRUE);
    }
  }  
   
  else if (creatures[cindex].position[0] == -80)
  {
    if(seg_intersect(creatures[cindex].position, v, bound[1][0], bound[2][0]))
    {
      *bound1 = 1; *bound2 = 2;
      return(TRUE);
    }
  }
  

  else if (creatures[cindex].position[2] == 80)
  {
    if(seg_intersect(creatures[cindex].position, v, bound[0][0], bound[1][0]))
    {
      *bound1 = 0; *bound2 = 1;
      return(TRUE);
    }
  }
  

  else if (creatures[cindex].position[2] == -80)
  {
    if(seg_intersect(creatures[cindex].position, v, bound[2][0], bound[3][0]))
    {
      *bound1 = 2; *bound2 = 3;
      return(TRUE);
    }
  }
  

  return(FALSE);
}



/* move the chip so it's within the center border of hole */
move_chip_within(VECTOR drop_point)
{
  VEC_NORM(drop_point);
  VEC_MULT(drop_point, 4.0);
}

/* is the chip being dropped intersecting a hole, goal, or another chip? 
   this function changes the value of the final destination if so */
find_chipblocks(CHIPS *chip, VECTOR drop_point)
{
  int i, j;
  int moved;
  VECTOR v, z;
  float dist;
  int chipi=5, chipj=5; /* who we are (don't know yet) */
  int count=0; /* makes sure we don't get stuck in a loop */
  
  /* check the hole first */
  do  {
    /* v represents the edge of the chip closest to the center */
    VEC_COPY(v, drop_point);
    v[1] = 0;
    dist = VEC_LEN(v);
    VEC_NORM(v);
    VEC_MULT(v, dist - .8);
    
    moved = FALSE;
    
    if (in_polygon(v, edge_border))
    {
      /* z represents the edge of the chip furthest from the center */
      VEC_COPY(z, drop_point);
      v[1] = 0;
      dist = VEC_LEN(z);
      VEC_NORM(z);
      VEC_MULT(z, dist + .8);

      if (in_polygon(z, center_border))
      {
	drop_point[1] = -3;
      }
      else
      {
	/* otherwise move it over so it's in the center */
	move_chip_within(drop_point);
	drop_point[1] = -3;
	moved = TRUE;
      }
    }
  } while(moved != FALSE);
  
  
  
  /* now check other chips */
  do {
    moved = FALSE;
    
    for (i=0; i<num_players; i++)
      for (j=0; j<4; j++)
      {
	if(((dist = VEC_DIST(drop_point, chips[i][j]->position)) < 1.4) &&
	   (i != chipi) && (j != chipj))
	{
	  if ((dist == 0) ||   
	      (VEC_DIST(drop_point, chips[i][j]->final_position) == 0))
	    /* we're that chip */
	  { 
	    chipi = i; chipj = j; 
	  }
	  dist -= .7; /* this gives us how much we have to move it */
	  VEC_COPY(v, drop_point);
	  VEC_NORM(v);  /* this is the direction we're moving it */
	  if ((v[0] == 0) && (v[2] == 0)) /* almost completely at the center */
	    VEC_COPY(v, chips[i][j]->position);  /* go opposite direction from
						    chip */
	  VEC_MULT(v, dist);
	  if (sqrt(SQR(drop_point[0]) + SQR(drop_point[2])) > 2) 
	    /* unless close to the center, move towards the center... 
	       so they all don't gather there */
	    VEC_NEG(v);
	  v[1] = 0;  /* moving it on a flat plane */
	  VEC_ADD(drop_point, v, drop_point);
	  moved = TRUE;
	}
      }
    count++;  /* after five tries, just let them intersect! */
  }  while((moved == TRUE) && (count < 5)); 
  
    
}

/* see if the creature is hitting the goalpost */
intersect_creature(CREATURE *creature, int goali, int goalj)
{
  int dir1, dir2, dir3;
  VECTOR correction;
  VECTOR temp, head, beak, tail;
  VECTOR v;
  TRI_LIST *tri;
  int i;

  /* find the locations of the major body parts */
  VEC_COPY(temp, looking);
  VEC_MULT(temp, -4);  /* beak is -4 from center of creature */
  mat_rot_apply(temp, temp, creature->beak.yrot_mat);
  VEC_ADD(beak, creature->position, temp);
  beak[1] = 0;

  VEC_COPY(temp, looking);
  VEC_MULT(temp, -4);  /* head is -4 from center of creature */
  VEC_ADD(head, creature->position, temp);
  head[1] = 0;
  
  VEC_COPY(temp, looking);
  VEC_MULT(temp, 3);  /* tail is 3 from center of creature */
  VEC_ADD(tail, creature->position, temp);
  tail[1] = 0;
  
  /* check intersections with bottom half of the goalpost */
  tri = first_triangle_intersect[goali*2 + goalj];
  for (i=0; i<7; i++)
  {
    if (intersect_tl(tri->t->v[0], tri->t->v[1], tri->t->v[2], 
		     head, tail) ||
	intersect_tl(tri->t->v[0], tri->t->v[1], tri->t->v[2], 
		     beak, tail))
    {
/*      VEC_SUB(v, creature->position, goal_start[i/2][ODD(i)]);
      VEC_NORM(v);
      v[1] = 0;
      VEC_ADD(creature->position, creature->position, v); */
      return(TRUE);
    }
    
    tri = tri->next;
  }

  
  return(FALSE);
}



/* loop through the set of goalposts, seeing if any intersect the line
   of sight. if so, make that goal wireframe. also check if other creatures 
   are in the way of movement */
find_viewblocks()
{
  int i,j,k;
  short front, back, center;
  int return_val = FALSE;
  VECTOR eye;
  VECTOR x, y;
  float facing;
  VECTOR temp, look, v;
  VECTOR creature_tip;
  VECTOR wing1_front, wing1_back, wing2_front, wing2_back;
  VECTOR wing1, wing2;
  TRI_LIST *tri, *tri1;
  PLAYER player;
  
  VEC_COPY(look, looking); /* direction of the camera */
  VEC_MULT(look, 30);  /* it's 30 units away */
  VEC_ADD(eye, look, creatures[me].position);

  VEC_COPY(temp, looking);
  VEC_MULT(temp, -4);  /* beak is 4 from center of creature */
  VEC_ADD(creature_tip, creatures[me].position, temp);



  VEC_SET(temp, looking[2], 0, -looking[0]); /* perpendicular */
  VEC_MULT(temp, 5);
  VEC_ADD(wing1, creatures[me].position, temp);
  VEC_NEG(temp);
  VEC_ADD(wing2, creatures[me].position, temp);

  VEC_SET(temp, looking[2], 0, -looking[0]); /* perpendicular */
  VEC_MULT(temp, 5);
  VEC_ADD(wing1_front, creatures[me].position, temp);
  
  VEC_ADD(wing1_back, creatures[me].position, temp);

  VEC_COPY(temp, looking);
  VEC_MULT(temp, 2);  /* width */
  VEC_ADD(wing1_front, wing1_front, temp);
  VEC_NEG(temp);
  VEC_ADD(wing1_back, wing1_back, temp);

  VEC_SET(temp, -looking[2], 0, looking[0]); /* other perpendicular */
  VEC_MULT(temp, 5);
  VEC_ADD(wing2_front, creatures[me].position, temp);
  
  VEC_ADD(wing2_back, creatures[me].position, temp);

  VEC_COPY(temp, looking);
  VEC_MULT(temp, 2);  /* width */
  VEC_ADD(wing2_front, wing2_front, temp);
  VEC_NEG(temp);
  VEC_ADD(wing2_back, wing2_back, temp);

  
  /* if we're close, check intersections with bottom 3rd of goalpost */
  for(i=0; i<3; i++)
    for(j=0; j<2; j++)
    {
      if ((VEC_DIST_SQR(goal_bounds[i][j][0], creature_tip) < 900) &&
	  (first_triangle[i*2 + j]->wireframe == FALSE));
      {
	/* check intersections with bottom third of the goalpost */
	tri = first_triangle_intersect[i*2 + j];
	tri1 = first_triangle_intersect[i*2 + j];
	for (k=0; k<7; k++)
	{
	  /* check if wing hit the post */
	  if  ((front = intersect_tl(tri1->t->v[0], tri1->t->v[1], 
				     tri1->t->v[2], 
				     wing1_front, wing2_front)) ||
	       (back = intersect_tl(tri1->t->v[0], tri1->t->v[1], 
				     tri1->t->v[2], 
				     wing1_back, wing2_back)) ||
	       (center = intersect_tl(tri1->t->v[0], tri1->t->v[1], 
				      tri1->t->v[2], 
				      wing1, wing2)))
	  {
	    return_val = TRUE;  /* wing hit the post */

	    /* find out which wing hit the post where */
	    if(front)
	    {
	      if (intersect_tl(tri1->t->v[0], tri1->t->v[1], 
			       tri1->t->v[2], wing1_front,
			       creatures[me].position)) 
	      {
		VEC_SUB(v, wing1_front, creatures[me].position);
		VEC_MULT(v, -.13);
		v[1] = 0; 
		VEC_ADD(creatures[me].position, creatures[me].position, v);
	      }
	      else/* if (intersect_tl(tri1->t->v[0], tri1->t->v[1], 
			       tri1->t->v[2], wing2_front,
			       creatures[me].position)) */
	      {
		VEC_SUB(v, wing2_front, creatures[me].position);
		VEC_MULT(v, -.13);
		v[1] = 0; 
		VEC_ADD(creatures[me].position, creatures[me].position, v);
	      }
	    }
	    else if(back)
	    {
	      if (intersect_tl(tri1->t->v[0], tri1->t->v[1], 
			       tri1->t->v[2], wing1_back,
			       creatures[me].position)) 
	      {
		VEC_SUB(v, wing1_back, creatures[me].position);
		VEC_MULT(v, -.13);
		v[1] = 0; 
		VEC_ADD(creatures[me].position, creatures[me].position, v);
	      }
	      else/* if (intersect_tl(tri1->t->v[0], tri1->t->v[1], 
			       tri1->t->v[2], wing1_back,
			       creatures[me].position))  */
	      {
		VEC_SUB(v, wing2_back, creatures[me].position);
		VEC_MULT(v, -.13);
		v[1] = 0; 
		VEC_ADD(creatures[me].position, creatures[me].position, v);
	      }
	    }
	    else if(center)
	    {
	      if (intersect_tl(tri1->t->v[0], tri1->t->v[1], 
			       tri1->t->v[2], wing1,
			       creatures[me].position)) 
	      {
		VEC_SUB(v, wing1, creatures[me].position);
		VEC_MULT(v, -.13);
		v[1] = 0; 
		VEC_ADD(creatures[me].position, creatures[me].position, v);
	      }
	      else/* if (intersect_tl(tri1->t->v[0], tri1->t->v[1], 
			       tri1->t->v[2], wing2,
			       creatures[me].position))  */
	      {
		VEC_SUB(v, wing2, creatures[me].position);
		VEC_MULT(v, -.13);
		v[1] = 0; 
		VEC_ADD(creatures[me].position, creatures[me].position, v);
	      }
	    }
	  }
	  else if (/*(k!=6)*/ (return_val == FALSE) && 
		   intersect_tl(tri->t->v[0], tri->t->v[1], tri->t->v[2], 
				creature_tip, eye))
	  {
	    k = 7;  /* stop the loop! */
	    if (!intersect_creature(&creatures[me], i, j)) 
	    { 
	      first_triangle[i*2 + j]->wireframe = TRUE;
	    }
	    else
	      return_val = TRUE;  /* creature hit the post */
	  }

	  tri = tri->next;
	  tri1 = tri1->next;
	}
      }	
    }

  /* other creatures */
  for (i=0; i<num_players; i++)
  {
    if ((i != me) && creatures[i].alive)
    {
      if (VEC_DIST(creatures[i].position, creatures[me].position) < 8)
      {
	return_val = TRUE;
	VEC_SUB(v, creatures[me].position, creatures[i].position);
	v[1] = 0;
	VEC_MULT(v, .25);
	VEC_ADD(creatures[me].position, creatures[me].position, v);
      }
    }
  }

  
  return(return_val);
}










