/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
/*
 * mx3.c: routines for manipulating 3x3 matrices
 * These matrices are commonly used to represent a 2-D perspective mapping.
 * Points are represented by row vectors and planes by columns.
 */

static char rcsid[] = "$Header: mx3.c,v 3.1 90/10/11 22:36:08 ph Locked $";
#include <math.h>

#include "mx3.h"
#include "det.h"

#define mx3(subr) mx3f_##subr
#define FLT float

/*-------------------- generic matrix operations --------------------*/

void mx3(identity)(a)
register FLT a[3][3];
{
    a[0][0] = 1.; a[0][1] = 0.; a[0][2] = 0.;
    a[1][0] = 0.; a[1][1] = 1.; a[1][2] = 0.;
    a[2][0] = 0.; a[2][1] = 0.; a[2][2] = 1.;
}

/* mx3_mul: matrix multiply: c = a*b */

void mx3(mul)(a, b, c)
register FLT a[3][3], b[3][3], c[3][3];
{
    c[0][0] = a[0][0]*b[0][0] + a[0][1]*b[1][0] + a[0][2]*b[2][0];
    c[0][1] = a[0][0]*b[0][1] + a[0][1]*b[1][1] + a[0][2]*b[2][1];
    c[0][2] = a[0][0]*b[0][2] + a[0][1]*b[1][2] + a[0][2]*b[2][2];
    c[1][0] = a[1][0]*b[0][0] + a[1][1]*b[1][0] + a[1][2]*b[2][0];
    c[1][1] = a[1][0]*b[0][1] + a[1][1]*b[1][1] + a[1][2]*b[2][1];
    c[1][2] = a[1][0]*b[0][2] + a[1][1]*b[1][2] + a[1][2]*b[2][2];
    c[2][0] = a[2][0]*b[0][0] + a[2][1]*b[1][0] + a[2][2]*b[2][0];
    c[2][1] = a[2][0]*b[0][1] + a[2][1]*b[1][1] + a[2][2]*b[2][1];
    c[2][2] = a[2][0]*b[0][2] + a[2][1]*b[1][2] + a[2][2]*b[2][2];
}

/* mx3_adjoint: b = adjoint(a), returns determinant(a) */

FLT mx3(adjoint)(a, b)
register FLT a[3][3], b[3][3];
{
    b[0][0] = DET2(a[1][1], a[1][2], a[2][1], a[2][2]);
    b[1][0] = DET2(a[1][2], a[1][0], a[2][2], a[2][0]);
    b[2][0] = DET2(a[1][0], a[1][1], a[2][0], a[2][1]);
    b[0][1] = DET2(a[2][1], a[2][2], a[0][1], a[0][2]);
    b[1][1] = DET2(a[2][2], a[2][0], a[0][2], a[0][0]);
    b[2][1] = DET2(a[2][0], a[2][1], a[0][0], a[0][1]);
    b[0][2] = DET2(a[0][1], a[0][2], a[1][1], a[1][2]);
    b[1][2] = DET2(a[0][2], a[0][0], a[1][2], a[1][0]);
    b[2][2] = DET2(a[0][0], a[0][1], a[1][0], a[1][1]);
    return a[0][0]*b[0][0] + a[0][1]*b[0][1] + a[0][2]*b[0][2];
}

void mx3(print)(str, a)
char *str;
FLT a[3][3];
{
    int i;

    printf("%s:\n", str);
    for (i=0; i<3; i++)
	printf("  %9.4f %9.4f %9.4f\n", a[i][0], a[i][1], a[i][2]);
}

/*-------------------- 3-D transformation operations --------------------*/

/* mx3_transform: transform point p by matrix a: q = p*a */

void mx3(transform)(p, a, q)
register FLT p[3], q[3];
register FLT a[3][3];
{
    q[0] = p[0]*a[0][0] + p[1]*a[1][0] + p[2]*a[2][0];
    q[1] = p[0]*a[0][1] + p[1]*a[1][1] + p[2]*a[2][1];
    q[2] = p[0]*a[0][2] + p[1]*a[1][2] + p[2]*a[2][2];
}

/*--------------- 2-D homogeneous transformation operations ---------------*/

/* mx3_transform_div: transform p by matrix a: q = p*a, doing homog. divide */

FLT mx3(transform_div)(p, a, q)
register FLT p[3], q[2];
register FLT a[3][3];
{
    FLT q2;

    q2 = p[0]*a[0][2] + p[1]*a[1][2] + p[2]*a[2][2];
    if (q2!=0.) {
	q[0] = (p[0]*a[0][0] + p[1]*a[1][0] + p[2]*a[2][0]) / q2;
	q[1] = (p[0]*a[0][1] + p[1]*a[1][1] + p[2]*a[2][1]) / q2;
    }
    return q2;
}

/* mx3_translate_mat: create matrix to translate by (tx,ty) */

void mx3(translate_mat)(tx, ty, m)
FLT tx, ty;
register FLT m[3][3];
{
    m[0][0] = 1.; m[0][1] = 0.; m[0][2] = 0.;
    m[1][0] = 0.; m[1][1] = 1.; m[1][2] = 0.;
    m[2][0] = tx; m[2][1] = ty; m[2][2] = 1.;
}

/* mx3_translate: premultiply m by translation of (tx,ty) */

void mx3(translate)(m, tx, ty)
register FLT m[3][3];
FLT tx, ty;
{
    m[2][0] += tx*m[0][0]+ty*m[1][0];
    m[2][1] += tx*m[0][1]+ty*m[1][1];
    m[2][2] += tx*m[0][2]+ty*m[1][2];
}

/* mx3_scale_mat: create matrix to scale by (sx,sy) */

void mx3(scale_mat)(sx, sy, m)
FLT sx, sy;
register FLT m[3][3];
{
    m[0][0] = sx; m[0][1] = 0.; m[0][2] = 0.;
    m[1][0] = 0.; m[1][1] = sy; m[1][2] = 0.;
    m[2][0] = 0.; m[2][1] = 0.; m[2][2] = 1.;
}

/* mx3_scale: premultiply m by scale of (sx,sy) */

void mx3(scale)(m, sx, sy)
register FLT m[3][3];
FLT sx, sy;
{
    m[0][0] *= sx; m[0][1] *= sx; m[0][2] *= sx;
    m[1][0] *= sy; m[1][1] *= sy; m[1][2] *= sy;
}

/* mx3_rotate_mat: create matrix to rotate angle radians */

void mx3(rotate_mat)(angle, m)
FLT angle;
register FLT m[3][3];
{
    FLT c, s;
    
    c = cos(angle);
    s = sin(angle);
    m[0][0] = c;  m[0][1] = s;  m[0][2] = 0.;
    m[1][0] = -s; m[1][1] = c;  m[1][2] = 0.;
    m[2][0] = 0.; m[2][1] = 0.; m[2][2] = 1.;
}

/* mx3_rotate: premultiply a rotation of angle radians */

void mx3(rotate)(m, angle)
FLT m[3][3];
FLT angle;
{
    FLT c, s, t;
    
    c = cos(angle);
    s = sin(angle);
    t = c*m[0][0]+s*m[1][0]; m[1][0] = c*m[1][0]-s*m[0][0]; m[0][0] = t;
    t = c*m[0][1]+s*m[1][1]; m[1][1] = c*m[1][1]-s*m[0][1]; m[0][1] = t;
    t = c*m[0][2]+s*m[1][2]; m[1][2] = c*m[1][2]-s*m[0][2]; m[0][2] = t;
}
