/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
#include <stdio.h>
#include <bstring.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <signal.h>

#include "net.h"
#include "config.h"
#include "opcodes.h"
#include "packet.h"
#include "world.h"
#include "server.h"
#include "creatures.h"

static NET_SELECT* net_select;
static int fd_server;

static SERVER_STATE server_state[MAX_PLAYERS];

static int game_in_progress;
static int handling_alarm_signals;
static int alarm_triggered;

short player_alive[MAX_PLAYERS];

extern WORLD world;

static server_state_init()
{
  int i;

  game_in_progress = 0;
  handling_alarm_signals = 0;
  alarm_triggered = 0;

  for (i = 0; i < MAX_PLAYERS; i++)
    server_state[i].mode = PLAYER_UNUSED;
}

static void server_alarm_handler(int sig, int code, struct sigcontext *sc)
{
  alarm_triggered = 1;
}

static void server_alarm_init()
{
  struct itimerval value;
  struct itimerval ovalue;
  struct sigaction act;

  handling_alarm_signals = 1;

  value.it_interval.tv_sec = 0;
  value.it_interval.tv_usec = 1000000/TICKS_PER_SECOND;
  value.it_value.tv_sec = 0;
  value.it_value.tv_usec = 1000000/TICKS_PER_SECOND;
  
  setitimer(ITIMER_REAL, &value, &ovalue);
  
  bzero(&act, sizeof(struct sigaction));
  act.sa_handler = server_alarm_handler;
  
  sigaction(SIGALRM, &act, NULL);
}

static server_add_player(int fd, struct sockaddr_in* addr)
{
  int i;
  
  i = 0;
  while (i < MAX_PLAYERS && server_state[i].mode != PLAYER_UNUSED) 
  {
    /* already joined */
    if (bcmp(&server_state[i].addr, addr, sizeof(struct sockaddr_in)) == 0)
      return;
    i++;
  }
  
  if (i == MAX_PLAYERS)
  {
    server_opcode_send(fd, addr, OP_GAME_IS_FULL);
    return;
  }
  
/*  if (game_in_progress) 
  {
    server_state[i].mode = PLAYER_WAITING;
    server_opcode_send(fd, addr, OP_GAME_IN_PROGRESS);
  } */
  
  else
  {
    if (i==0)
      world_init(); 
    world.num_players = i+1;
    world.new_player = i;
    server_state[i].mode = PLAYER_ACTIVE;
    world_send_update(fd, addr);
  }

  bcopy(addr, &server_state[i].addr, sizeof(struct sockaddr_in));

  if (!handling_alarm_signals)
    server_alarm_init();
}

static void server_read_handler(int fd)
{
  struct sockaddr_in addr;
  CLIENT_PACKET client_packet;
  CHIP_INFO info;
  int i, alive=0;
  
  client_packet_recv(fd, &addr, &client_packet);
  
  switch (client_packet.opcode)
  {
  case OP_JOIN_REQUEST :
    server_add_player(fd, &addr);
    break;
  case OP_PLAYER_UPDATE :
    player_receive_update(&client_packet.item.player);
    world_send_update(fd, &addr);

    player_alive[client_packet.item.player.index] = 
      client_packet.item.player.alive;
    for (i=0; i<world.num_players; i++)
      if (player_alive[i]) alive++;
    if (alive == 0)
    {
      printf("tagh: (server) Terminated.\n");
      exit(0);
    } 
    break;
  case OP_GAME_UPDATE :
    for(i=0; i<world.num_players; i++)
      world.creat_it[i] = C_NOTIT;
    world.creat_it[client_packet.item.it] = C_IT;
    world_send_update(fd, &addr); 
    break; 
  case OP_CHIP_UPDATE :
    info = client_packet.item.chip;
    VEC_COPY(world.chip_position[info.indexi][info.indexj], info.position);
    VEC_COPY(world.chip_final_position[info.indexi][info.indexj], 
	     info.final_position);
    world.chip_yrot[info.indexi][info.indexj] = info.yrot;
    world.chip_status[info.indexi][info.indexj] = info.status;
    world_send_update(fd, &addr); 
    break;
  case OP_END_GAME :
    world_renew();
    world_send_update(fd, &addr); 
    break;
  }

}

int server_init()
{
  int pid;			/* server's process ID */
  int fd;			/* server's file descriptor */

  server_state_init();

  pid = fork();
  if (pid == -1)
    return SERVER_ERROR;

  if (pid != 0)
  {
    world.num_players = 1;
    world.creat_it[1] = C_IT;
    return SERVER_OK;
  }
  
  fd_server = net_open(SERVER_PORT);
  if (fd_server == NET_ERROR)
  {
    fprintf(stderr, "tagh: Server port already in use.\n");
    exit(-1);
  }
  
  net_select = net_select_new();
  if (net_select == NULL)
  {
    fprintf(stderr, "tagh: Can't initialize server.\n");
    exit(-1);
  }
  
  net_select_timeout(net_select, SERVER_TIMEOUT, 0);
  net_select_set(net_select, fd_server, server_read_handler);

  while (net_select_loop(net_select) == NET_ERROR)
    if (alarm_triggered)
    {
      world_update();
      alarm_triggered = 0;
    }
  
  fprintf(stderr, "tagh: Server timed out.\n");
  
  exit(-1);
}
