/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
#include <stdio.h>
#include <math.h>
#include <malloc.h>

#include "vroom.h"
#include "track.h"
#include "car.h"
#include "body.h"

#define	FOLLOW_OK		0
#define	FOLLOW_NEXT		1
#define	FOLLOW_WIPE_OUT_LEFT	2
#define	FOLLOW_WIPE_OUT_RIGHT	3

#define	MID_RADIUS(t)	( (t)->c[0] )
#define	INNER_RADIUS(t)	( (t)->c[1] )
#define	OUTER_RADIUS(t)	( (t)->c[2] )
#define	PIVOT_X(t)	( (t)->c[3] )
#define	PIVOT_Y(t)	( (t)->c[4] )
#define	TURN_ANGLE(t)	( (t)->c[5] )

/* BEGIN PROTOTYPES -S curved.c */
static void     addedLeftCurvedTrack( Track *track ) ;
static void     addedRightCurvedTrack( Track *track ) ;
static void     addToLeftCurvedTrack( Track *old, Track *new ) ;
static void     addToRightCurvedTrack( Track *old, Track *new ) ;
static float    curvedLaneWidth( Track *track, float s, float lane1,
                    float lane2 ) ;
static void     drawLeftCurvedOutline( Track *track, int side ) ;
static void     drawLeftCurvedTrack( Track *track, int lod ) ;
static void     drawLeftCurvedTrackMarkers( Track *track, int lod ) ;
static void     drawRightCurvedOutline( Track *track, int side ) ;
static void     drawRightCurvedTrack( Track *track, int lod ) ;
static void     drawRightCurvedTrackMarkers( Track *track, int lod ) ;
static void     followLeftCurvedTrack( Track *track, Car *car, float dTime ) ;
static void     followRightCurvedTrack( Track *track, Car *car, float dTime ) ;
static void     freeCurvedTrack( Track *track ) ;
static float    getLeftCurvedTrackLength( Track *track, float lane ) ;
static float    getRightCurvedTrackLength( Track *track, float lane ) ;
static float    topSpeedLeftCurvedTrack( Track *track, float s, float l ) ;
static float    topSpeedRightCurvedTrack( Track *track, float s, float l ) ;
/* END PROTOTYPES -S curved.c */

		

/*------------------------------------------------------------------------------
 * Follow a curved piece of track.
 *----------------------------------------------------------------------------*/
static int
followCurvedTrack(
	Track	*track,
	Car	*car,
	float	dTime
	)
{
	int	doNext = 0 ;
	int	laneLo ;
	int	laneHi ;
	float	length ;
	float	radius ;
	float	theta ;
	float	c ;
	float	s ;
	float	eTime ;
	float	newS ;
	float	partialLane ;
	Track	*nextTrack ;

	if( car->lane < 1.0f || car->lane > (float)track->lanes )
	{
		if( car->lane < 1.0f )
		{
			return( FOLLOW_WIPE_OUT_LEFT ) ;
		}
		else
		{
			return( FOLLOW_WIPE_OUT_RIGHT ) ;
		}
	}
	else
	{
		laneLo = (int)floorf( car->lane ) ;
		laneHi = (int)ceilf( car->lane ) ;
		partialLane = car->lane - laneLo ;
		length = partialLane * track->length[laneLo-1] +
			( 1.0f - partialLane ) * track->length[laneHi-1] ;
		newS = car->s + car->speed * dTime / length ;
		if( newS >= 1.0f )
		{
			eTime = ( 1.0f - car->s ) * length /
				car->speed ;
			dTime -= eTime ;
			car->s = 0.0f ;
			car->track = track->nextTrack ;
			incrementLap( car, dTime ) ;
			car->track->enter( car->track, car ) ;
			doNext = 1 ;
		}
		else
		{
			eTime = dTime ;
			car->s = newS ;
		}
		if( car->status & CHANGE_LANE_RIGHT )
		{
			partialLane += eTime * LANE_CHANGE_SPEED ;
			if( partialLane > 1.0f )
			{
				car->status &= ~CHANGE_LANE_RIGHT ;
				car->lane = laneLo + 1.0f ;
			}
			else
			{
				car->lane = laneLo + partialLane ;
			}
		}
		else if( car->status & CHANGE_LANE_LEFT )
		{
			partialLane = laneHi - car->lane ;
			partialLane += eTime * LANE_CHANGE_SPEED ;
			if( partialLane > 1.0f )
			{
				car->status &= ~CHANGE_LANE_LEFT ;
				car->lane = laneHi - 1.0f ;
			}
			else
			{
				car->lane = laneHi - partialLane ;
			}
		}
		if( doNext )
		{
			car->track->follow( car->track, car, dTime ) ;
			return( FOLLOW_NEXT ) ;
		}
	}

	return( FOLLOW_OK ) ;
}



/*------------------------------------------------------------------------------
 * Follow a curved piece of track.
 *----------------------------------------------------------------------------*/
static void
followLeftCurvedTrack(
	Track	*track,
	Car	*car,
	float	dTime
	)
{
	float	wvx ;
	float	wvy ;
	float	radius ;
	float	theta ;
	float	c ;
	float	s ;
	float	force ;

	switch( followCurvedTrack( track, car, dTime ) )
	{
		case FOLLOW_OK :
			radius = MID_RADIUS( track ) + LANE_WIDTH *
				( car->lane - track->midLane ) ;
			theta = track->theta + car->s * TURN_ANGLE( track ) ;
			c = cosf( theta ) ;
			s = sinf( theta ) ;
			car->x = PIVOT_X( track ) + radius * c ;
			car->y = PIVOT_Y( track ) + radius * s ;
			car->theta = theta ;
			force = centrifugalForce( car->speed, radius ) ;
			if( force >= 1.0f )
			{
				wvx = cosf( theta + (float)M_PI * 0.3333f ) ;
				wvy = sinf( theta + (float)M_PI * 0.3333f ) ;
				wipeOutCar( car, wvx, wvy, 1, WIPE_OUT_SPEED ) ;
			}
			else if( force > 0.0f )
			{
				startSkid( car, force, dTime ) ;
			}
			else
			{
				if( car->status & CAR_SKID )
				{
					stopSkid( car ) ;
				}
			}
			break ;

		case FOLLOW_NEXT :
			break ;

		case FOLLOW_WIPE_OUT_LEFT :
			theta = track->theta + car->s * TURN_ANGLE( track ) ;
			wvx = cosf( theta + (float)M_PI * 0.8333f ) ;
			wvy = sinf( theta + (float)M_PI * 0.8333f ) ;
			wipeOutCar( car, wvx, wvy, -1, WIPE_OUT_OFF_ROAD ) ;
			break ;

		case FOLLOW_WIPE_OUT_RIGHT :
			theta = track->theta + car->s * TURN_ANGLE( track ) ;
			wvx = cosf( theta + (float)M_PI * 0.3333f ) ;
			wvy = sinf( theta + (float)M_PI * 0.3333f ) ;
			wipeOutCar( car, wvx, wvy, 1, WIPE_OUT_OFF_ROAD ) ;
			break ;
	}
}



/*------------------------------------------------------------------------------
 * Follow a curved piece of track.
 *----------------------------------------------------------------------------*/
static void
followRightCurvedTrack(
	Track	*track,
	Car	*car,
	float	dTime
	)
{
	float	wvx ;
	float	wvy ;
	float	radius ;
	float	theta ;
	float	c ;
	float	s ;
	float	force ;

	switch( followCurvedTrack( track, car, dTime ) )
	{
		case FOLLOW_OK :
			radius = MID_RADIUS( track ) - LANE_WIDTH *
				( car->lane - track->midLane ) ;
			theta = track->theta + car->s * TURN_ANGLE( track ) ;
			c = cosf( theta ) ;
			s = sinf( theta ) ;
			car->x = PIVOT_X( track ) - radius * c ;
			car->y = PIVOT_Y( track ) - radius * s ;
			car->theta = theta ;
			force = centrifugalForce( car->speed, radius ) ;
			if( force >= 1.0f )
			{
				wvx = cosf( theta + (float)M_PI * 0.6667f );
				wvy = sinf( theta + (float)M_PI * 0.6667f );
				wipeOutCar( car, wvx, wvy, 1, WIPE_OUT_SPEED ) ;
			}
			else if( force > 0.0f )
			{
				startSkid( car, -force, dTime ) ;
			}
			else
			{
				if( car->status & CAR_SKID )
				{
					stopSkid( car ) ;
				}
			}
			break ;

		case FOLLOW_NEXT :
			break ;

		case FOLLOW_WIPE_OUT_LEFT :
			theta = track->theta + car->s * TURN_ANGLE( track ) ;
			wvx = cosf( theta + (float)M_PI * 0.6667f ) ;
			wvy = sinf( theta + (float)M_PI * 0.6667f ) ;
			wipeOutCar( car, wvx, wvy, -1, WIPE_OUT_OFF_ROAD ) ;
			break ;

		case FOLLOW_WIPE_OUT_RIGHT :
			theta = track->theta + car->s * TURN_ANGLE( track ) ;
			wvx = cosf( theta + (float)M_PI * 0.1667f ) ;
			wvy = sinf( theta + (float)M_PI * 0.1667f ) ;
			wipeOutCar( car, wvx, wvy, 1, WIPE_OUT_OFF_ROAD ) ;
			break ;
	}
}



/*------------------------------------------------------------------------------
 * Initialize after adding track.
 *----------------------------------------------------------------------------*/
static void
addedLeftCurvedTrack(
	Track	*track
	)
{
	int	i ;
	float	radius ;
	float	a[2] ;
	float	b[2] ;
	float	r[2] ;
	float	d[2] ;
	float	cts ;
	float	sts ;
	float	f ;

	PIVOT_X( track ) = track->x - MID_RADIUS( track ) *
				cosf( track->theta ) ;
	PIVOT_Y( track ) = track->y - MID_RADIUS( track ) *
				sinf( track->theta ) ;

	/*
	 * Compute the corner points of the starting edge of the track.
	 */
	cts = cosf( track->theta ) ;
	sts = sinf( track->theta ) ;
	radius = INNER_RADIUS( track ) - SIDE_MARKER_WIDTH ;
	b[0] = PIVOT_X( track ) + radius * cts ;
	b[1] = PIVOT_Y( track ) + radius * sts ;
	radius = OUTER_RADIUS( track ) + SIDE_MARKER_WIDTH ;
	a[0] = PIVOT_X( track ) + radius * cts ;
	a[1] = PIVOT_Y( track ) + radius * sts ;
	/*
	 * Know that the center point of the track is going to be along
	 * the half-turning angle ray, so solve for the point along that
	 * ray that is equi-distant from the corner points.
	 */
	r[0] = cosf( track->theta + 0.5f * TURN_ANGLE( track ) ) ;
	r[1] = sinf( track->theta + 0.5f * TURN_ANGLE( track ) ) ;
	f = ( 0.5f * ( b[0]*b[0] - a[0]*a[0] + b[1]*b[1] - a[1]*a[1] ) +
		( a[0] - b[0] ) * PIVOT_X( track ) +
		( a[1] - b[1] ) * PIVOT_Y( track ) ) /
		( ( b[0] - a[0] ) * r[0] + ( b[1] - a[1] ) * r[1] ) ;
	track->center[0] = PIVOT_X( track ) + f * r[0] ;
	track->center[1] = PIVOT_Y( track ) + f * r[1] ;
	track->center[2] = 0.0f ;
	/*
	 * Compute the cull radius from the center point.
	 */
	radius = sqrtf( ( a[0]-track->center[0] ) * ( a[0]-track->center[0] ) +
			( a[1]-track->center[1] ) * ( a[1]-track->center[1] ) );
	track->cullDetail = trackDetail( radius ) ;

	for( i = 0 ; i < MAX_LOD ; i++ )
	{
		glNewList( track->drawList+i, GL_COMPILE ) ;
		drawLeftCurvedTrack( track, i ) ;
		glEndList() ;
		glNewList( track->drawMarkerList+i, GL_COMPILE ) ;
		drawLeftCurvedTrackMarkers( track, i ) ;
		glEndList() ;
	}

}



/*------------------------------------------------------------------------------
 * Initialize after adding track.
 *----------------------------------------------------------------------------*/
static void
addedRightCurvedTrack(
	Track	*track
	)
{
	int	i ;
	float	radius ;
	float	a[2] ;
	float	b[2] ;
	float	r[2] ;
	float	d[2] ;
	float	cts ;
	float	sts ;
	float	f ;

	PIVOT_X( track ) = track->x + MID_RADIUS( track ) *
				cosf( track->theta ) ;
	PIVOT_Y( track ) = track->y + MID_RADIUS( track ) *
				sinf( track->theta ) ;

	/*
	 * Compute the corner points of the starting edge of the track.
	 */
	cts = cosf( track->theta ) ;
	sts = sinf( track->theta ) ;
	radius = INNER_RADIUS( track ) - SIDE_MARKER_WIDTH ;
	b[0] = PIVOT_X( track ) - radius * cts ;
	b[1] = PIVOT_Y( track ) - radius * sts ;
	radius = OUTER_RADIUS( track ) + SIDE_MARKER_WIDTH ;
	a[0] = PIVOT_X( track ) - radius * cts ;
	a[1] = PIVOT_Y( track ) - radius * sts ;
	/*
	 * Know that the center point of the track is going to be along
	 * the half-turning angle ray, so solve for the point along that
	 * ray that is equi-distant from the corner points.
	 */
	r[0] = cosf( track->theta + 0.5f * TURN_ANGLE( track ) ) ;
	r[1] = sinf( track->theta + 0.5f * TURN_ANGLE( track ) ) ;
	f = ( 0.5f * ( b[0]*b[0] - a[0]*a[0] + b[1]*b[1] - a[1]*a[1] ) +
		( a[0] - b[0] ) * PIVOT_X( track ) +
		( a[1] - b[1] ) * PIVOT_Y( track ) ) /
		( ( b[0] - a[0] ) * r[0] + ( b[1] - a[1] ) * r[1] ) ;
	track->center[0] = PIVOT_X( track ) + f * r[0] ;
	track->center[1] = PIVOT_Y( track ) + f * r[1] ;
	track->center[2] = 0.0f ;
	/*
	 * Compute the cull radius from the center point.
	 */
	radius = sqrtf( ( a[0]-track->center[0] ) * ( a[0]-track->center[0] ) +
			( a[1]-track->center[1] ) * ( a[1]-track->center[1] ) );
	track->cullDetail = trackDetail( radius ) ;

	for( i = 0 ; i < MAX_LOD ; i++ )
	{
		glNewList( track->drawList+i, GL_COMPILE ) ;
		drawRightCurvedTrack( track, i ) ;
		glEndList() ;
		glNewList( track->drawMarkerList+i, GL_COMPILE ) ;
		drawRightCurvedTrackMarkers( track, i ) ;
		glEndList() ;
	}
}



/*------------------------------------------------------------------------------
 * Create a N lane piece of left curved track.
 *----------------------------------------------------------------------------*/
Track *
createLeftCurvedTrack(
	int	nLanes,
	float	radius,
	float	angle
	)
{
	int	i ;
	float	r ;
	Track	*track ;

	track = myMalloc( sizeof( Track ) ) ;

	track->lanes = nLanes ;
	track->type = L_CURVE_TRACK ;
	track->x = 0.0f ;
	track->y = 0.0f ;
	track->theta = 0.0f ;
	track->length = myMalloc( nLanes * sizeof( float ) ) ;
	r = radius - 0.5f * ( nLanes - 1 ) * LANE_WIDTH ;
	INNER_RADIUS( track ) = r - 0.5f * LANE_WIDTH ;
	angle *= M_PI / 180.0f ;
	TURN_ANGLE( track ) = angle ;
	for( i = 0 ; i < nLanes ; i++ )
	{
		track->length[i] = r * angle ;
		r += LANE_WIDTH ;
	}
	OUTER_RADIUS( track ) = r - 0.5f * LANE_WIDTH ;
	MID_RADIUS( track ) = radius ;
	track->midLane = 1.0f + 0.5f * ( nLanes - 1 ) ;
	track->follow = followLeftCurvedTrack ;
	track->add = addToLeftCurvedTrack ;
	track->added = addedLeftCurvedTrack ;
	track->topSpeed = topSpeedLeftCurvedTrack ;
	track->nextTrack = NULL ;
	track->drawList = glGenLists( MAX_LOD ) ;
	track->drawMarkerList = glGenLists( MAX_LOD ) ;
	track->startPosition = NULL ;
	track->endPosition = NULL ;
	track->drawOutline = drawLeftCurvedOutline ;
	track->trackLength = getLeftCurvedTrackLength ;
	track->enter = enterTrack ;
	track->free = freeCurvedTrack ;
	track->cullDetail = NULL ;
	track->laneWidth = curvedLaneWidth ;

	return( track ) ;
}



/*------------------------------------------------------------------------------
 * Create a N lane piece of right curved track.
 *----------------------------------------------------------------------------*/
Track *
createRightCurvedTrack(
	int	nLanes,
	float	radius,
	float	angle
	)
{
	int	i ;
	float	r ;
	Track	*track ;

	track = myMalloc( sizeof( Track ) ) ;

	track->lanes = nLanes ;
	track->type = R_CURVE_TRACK ;
	track->x = 0.0f ;
	track->y = 0.0f ;
	track->theta = 0.0f ;
	track->length = myMalloc( nLanes * sizeof( float ) ) ;
	r = radius - 0.5f * ( nLanes - 1 ) * LANE_WIDTH ;
	INNER_RADIUS( track ) = r - 0.5f * LANE_WIDTH ;
	angle *= M_PI / 180.0f ;
	TURN_ANGLE( track ) = -angle ;
	for( i = 0 ; i < nLanes ; i++ )
	{
		track->length[nLanes-i-1] = r * angle ;
		r += LANE_WIDTH ;
	}
	OUTER_RADIUS( track ) = r - 0.5f * LANE_WIDTH ;
	MID_RADIUS( track ) = radius ;
	track->midLane = 1.0f + 0.5f * ( nLanes - 1 ) ;
	track->follow = followRightCurvedTrack ;
	track->add = addToRightCurvedTrack ;
	track->added = addedRightCurvedTrack ;
	track->topSpeed = topSpeedRightCurvedTrack ;
	track->nextTrack = NULL ;
	track->drawList = glGenLists( MAX_LOD ) ;
	track->drawMarkerList = glGenLists( MAX_LOD ) ;
	track->startPosition = NULL ;
	track->endPosition = NULL ;
	track->drawOutline = drawRightCurvedOutline ;
	track->trackLength = getRightCurvedTrackLength ;
	track->enter = enterTrack ;
	track->free = freeCurvedTrack ;
	track->cullDetail = NULL ;
	track->laneWidth = curvedLaneWidth ;

	return( track ) ;
}



/*------------------------------------------------------------------------------
 * Add to a piece of curved track.
 *----------------------------------------------------------------------------*/
static void
addToLeftCurvedTrack(
	Track	*old,
	Track	*new
	)
{
	new->theta = old->theta + TURN_ANGLE( old ) ;
	new->x = PIVOT_X( old ) + MID_RADIUS( old ) * cosf( new->theta ) ;
	new->y = PIVOT_Y( old ) + MID_RADIUS( old ) * sinf( new->theta ) ;
	old->nextTrack = new ;
}



/*------------------------------------------------------------------------------
 * Add to a piece of curved track.
 *----------------------------------------------------------------------------*/
static void
addToRightCurvedTrack(
	Track	*old,
	Track	*new
	)
{
	new->theta = old->theta + TURN_ANGLE( old ) ;
	new->x = PIVOT_X( old ) - MID_RADIUS( old ) * cosf( new->theta ) ;
	new->y = PIVOT_Y( old ) - MID_RADIUS( old ) * sinf( new->theta ) ;
	old->nextTrack = new ;
}



/*------------------------------------------------------------------------------
 * Draw the left curved track piece.
 *----------------------------------------------------------------------------*/
static void
drawLeftCurvedTrack(
	Track	*track,
	int	lod
	)
{
	int	i ;
	int	n ;
	int	l ;
	GLfloat	t ;
	GLfloat	c ;
	GLfloat	s ;
	GLfloat	ri ;
	GLfloat	ro ;
	float	ta = TURN_ANGLE( track ) ;

	n = 16 - lod * 4 ;
	if( n < 4 )
	{
		n = 4 ;
	}

	for( l = 0 ; l < track->lanes ; l++ )
	{
		glBegin( GL_TRIANGLE_STRIP ) ;
		ri = INNER_RADIUS( track ) + 0.5f * ( LANE_WIDTH - RAIL_WIDTH )
			+ l * LANE_WIDTH ;
		ro = ri + RAIL_WIDTH ;
		for( i = 0 ; i < n ; i++ )
		{
			t = track->theta + ( (float)i / (float)( n-1 ) ) * ta ;
			c = cosf( t ) ;
			s = sinf( t ) ;
			glVertex2f( PIVOT_X( track ) + ri * c,
				    PIVOT_Y( track ) + ri * s ) ;
			glVertex2f( PIVOT_X( track ) + ro * c,
				    PIVOT_Y( track ) + ro * s ) ;
		}
		glEnd() ;
	}
}



/*------------------------------------------------------------------------------
 * Draw the right curved track piece.
 *----------------------------------------------------------------------------*/
static void
drawRightCurvedTrack(
	Track	*track,
	int	lod
	)
{
	int	i ;
	int	n ;
	int	l ;
	GLfloat	t ;
	GLfloat	c ;
	GLfloat	s ;
	GLfloat	ri ;
	GLfloat	ro ;
	float	ta = -TURN_ANGLE( track ) ;

	n = 16 - lod * 4 ;
	if( n < 4 )
	{
		n = 4 ;
	}
	for( l = 0 ; l < track->lanes ; l++ )
	{
		glBegin( GL_TRIANGLE_STRIP ) ;
		ri = INNER_RADIUS( track ) + 0.5f * ( LANE_WIDTH - RAIL_WIDTH )
			+ l * LANE_WIDTH ;
		ro = ri + RAIL_WIDTH ;
		for( i = 0 ; i < n ; i++ )
		{
			t = track->theta - ( (float)i / (float)( n-1 ) ) * ta ;
			c = cosf( t ) ;
			s = sinf( t ) ;
			glVertex2f( PIVOT_X( track ) - ro * c,
				    PIVOT_Y( track ) - ro * s ) ;
			glVertex2f( PIVOT_X( track ) - ri * c,
				    PIVOT_Y( track ) - ri * s ) ;
		}
		glEnd() ;
	}
}



/*------------------------------------------------------------------------------
 * Draw the left curved track markers.
 *----------------------------------------------------------------------------*/
static void
drawLeftCurvedTrackMarkers(
	Track	*track,
	int	lod
	)
{
	int	i ;
	int	n ;
	int	l ;
	GLfloat	t ;
	GLfloat	c ;
	GLfloat	s ;
	GLfloat	ri ;
	GLfloat	ro ;
	float	ta = TURN_ANGLE( track ) ;

	if( lod < MAX_LOD - 1 )
	{
		t = INNER_RADIUS( track ) * ta ;
		ri = INNER_RADIUS( track ) - SIDE_MARKER_WIDTH ;
		ro = INNER_RADIUS( track ) ;
		for( l = 0 ; l < 2 ; l++ )
		{
			n = (int)( t / SIDE_MARKER_LENGTH ) ;
			n += ( 4 - ( n % 4 ) ) % 4 ;
			glBegin( GL_QUADS ) ;
			for( i = 0 ; i < n ; i += 2 )
			{
				if( ( i % 4 ) == 0 )
				{
					glColor3f( 0.5f, 0.0f, 0.0f ) ;
				}
				else
				{
					glColor3f( 0.9f, 0.9f, 0.9f ) ;
				}
				t = track->theta + ( (float)i / (float)n ) * ta;
				c = cosf( t ) ;
				s = sinf( t ) ;
				glVertex2f( PIVOT_X( track ) + ri * c,
					    PIVOT_Y( track ) + ri * s ) ;
				glVertex2f( PIVOT_X( track ) + ro * c,
					    PIVOT_Y( track ) + ro * s ) ;
				t += ta / (float)n ;
				c = cosf( t ) ;
				s = sinf( t ) ;
				glVertex2f( PIVOT_X( track ) + ro * c,
					    PIVOT_Y( track ) + ro * s ) ;
				glVertex2f( PIVOT_X( track ) + ri * c,
					    PIVOT_Y( track ) + ri * s ) ;
			}
			glEnd() ;
			t = OUTER_RADIUS( track ) * ta ;
			ri = OUTER_RADIUS( track ) ;
			ro = OUTER_RADIUS( track ) + SIDE_MARKER_WIDTH ;
		}
	}
}



/*------------------------------------------------------------------------------
 * Draw the right curved track markers.
 *----------------------------------------------------------------------------*/
static void
drawRightCurvedTrackMarkers(
	Track	*track,
	int	lod
	)
{
	int	i ;
	int	n ;
	int	l ;
	GLfloat	t ;
	GLfloat	c ;
	GLfloat	s ;
	GLfloat	ri ;
	GLfloat	ro ;
	float	ta = -TURN_ANGLE( track ) ;

	if( lod < MAX_LOD - 1 )
	{
		t = INNER_RADIUS( track ) * ta ;
		ri = INNER_RADIUS( track ) - SIDE_MARKER_WIDTH ;
		ro = INNER_RADIUS( track ) ;
		for( l = 0 ; l < 2 ; l++ )
		{
			n = (int)( t / SIDE_MARKER_LENGTH ) ;
			n += ( 4 - ( n % 4 ) ) % 4 ;
			glBegin( GL_QUADS ) ;
			for( i = 0 ; i < n ; i += 2 )
			{
				if( ( i % 4 ) == 0 )
				{
					glColor3f( 0.5f, 0.0f, 0.0f ) ;
				}
				else
				{
					glColor3f( 0.9f, 0.9f, 0.9f ) ;
				}
				t = track->theta - ( (float)i / (float)n ) * ta;
				c = cosf( t ) ;
				s = sinf( t ) ;
				glVertex2f( PIVOT_X( track ) - ro * c,
					    PIVOT_Y( track ) - ro * s ) ;
				glVertex2f( PIVOT_X( track ) - ri * c,
					    PIVOT_Y( track ) - ri * s ) ;
				t -= ta / (float)n ;
				c = cosf( t ) ;
				s = sinf( t ) ;
				glVertex2f( PIVOT_X( track ) - ri * c,
					    PIVOT_Y( track ) - ri * s ) ;
				glVertex2f( PIVOT_X( track ) - ro * c,
					    PIVOT_Y( track ) - ro * s ) ;
			}
			glEnd() ;
			t = OUTER_RADIUS( track ) * ta ;
			ri = OUTER_RADIUS( track ) ;
			ro = OUTER_RADIUS( track ) + SIDE_MARKER_WIDTH ;
		}
	}
}



/*------------------------------------------------------------------------------
 * Draw the left curved track piece outline.
 *----------------------------------------------------------------------------*/
static void
drawLeftCurvedOutline(
	Track	*track,
	int	side
	)
{
	int	i ;
	int	n = 5 ;
	GLfloat	r ;
	GLfloat	t ;
	GLfloat	c ;
	GLfloat	s ;
	GLfloat	a[2] ;
	float	ta = TURN_ANGLE( track ) ;

	r = ( side == TRACK_LEFT_SIDE ) ? INNER_RADIUS( track ) :
					OUTER_RADIUS( track ) ;
	for( i = 1 ; i < n ; i++ )
	{
		t = ( (float)i / (float)( n - 1 ) ) * ta ;
		c = cosf( track->theta + t ) ;
		s = sinf( track->theta + t ) ;
		a[0] = PIVOT_X( track ) + r * c ;
		a[1] = PIVOT_Y( track ) + r * s ;
		glVertex2fv( a ) ;
	}
}



/*------------------------------------------------------------------------------
 * Draw the right curved track piece outline.
 *----------------------------------------------------------------------------*/
static void
drawRightCurvedOutline(
	Track	*track,
	int	side
	)
{
	int	i ;
	int	n = 5 ;
	GLfloat	r ;
	GLfloat	t ;
	GLfloat	c ;
	GLfloat	s ;
	GLfloat	a[2] ;
	float	ta = -TURN_ANGLE( track ) ;

	r = ( side == TRACK_RIGHT_SIDE ) ? INNER_RADIUS( track ) :
					OUTER_RADIUS( track ) ;

	for( i = 1 ; i < n ; i++ )
	{
		t = ( (float)i / (float)( n - 1 ) ) * ta ;
		c = cosf( track->theta - t ) ;
		s = sinf( track->theta - t ) ;
		a[0] = PIVOT_X( track ) - r * c ;
		a[1] = PIVOT_Y( track ) - r * s ;
		glVertex2fv( a ) ;
	}
}



/*------------------------------------------------------------------------------
 * Return the top speed of a left curved track.
 *----------------------------------------------------------------------------*/
static float
topSpeedLeftCurvedTrack(
	Track	*track,
	float	s,
	float	l
	)
{
	float	radius ;

	radius = MID_RADIUS( track ) + LANE_WIDTH * ( l - track->midLane ) ;

	return( maxSpeed( radius ) ) ;
}



/*------------------------------------------------------------------------------
 * Return the top speed of a right curved track.
 *----------------------------------------------------------------------------*/
static float
topSpeedRightCurvedTrack(
	Track	*track,
	float	s,
	float	l
	)
{
	float	radius ;

	radius = MID_RADIUS( track ) - LANE_WIDTH * ( l - track->midLane ) ;

	return( maxSpeed( radius ) ) ;
}



/*------------------------------------------------------------------------------
 * Free up memory used by a piece of track.
 *----------------------------------------------------------------------------*/
static void
freeCurvedTrack(
	Track	*track
	)
{
	if( track )
	{
		free( track->length ) ;
		free( track ) ;
	}
}



/*------------------------------------------------------------------------------
 * Return the length of track at the given lane.
 *----------------------------------------------------------------------------*/
static float
getLeftCurvedTrackLength(
	Track	*track,
	float	lane
	)
{
	float	radius ;

	radius = MID_RADIUS( track ) + LANE_WIDTH * ( lane - track->midLane ) ;

	return( radius * TURN_ANGLE( track ) ) ;
}



/*------------------------------------------------------------------------------
 * Return the length of track at the given lane.
 *----------------------------------------------------------------------------*/
static float
getRightCurvedTrackLength(
	Track	*track,
	float	lane
	)
{
	float	radius ;

	radius = MID_RADIUS( track ) - LANE_WIDTH * ( lane - track->midLane ) ;

	return( -radius * TURN_ANGLE( track ) ) ;
}



/*------------------------------------------------------------------------------
 * Return the location of a point on the track.
 *----------------------------------------------------------------------------*/
static float
curvedLaneWidth(
	Track	*track,
	float	s,
	float	lane1,
	float	lane2
	)
{
	if( lane1 < 1.0f )
	{
		lane1 = 1.0f ;
	}
	else if( lane1 > (float)track->lanes )
	{
		lane1 = (float)track->lanes ;
	}
	if( lane2 < 1.0f )
	{
		lane2 = 1.0f ;
	}
	else if( lane2 > (float)track->lanes )
	{
		lane2 = (float)track->lanes ;
	}

	lane1 = ( lane1 - lane2 ) * LANE_WIDTH ;
	return( ABSFUNC( lane1 ) ) ;
}
