/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include <malloc.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/prctl.h>
#include <invent.h>
#include <audio.h>
#include <audiofile.h>
#include <signal.h>
#include <stropts.h>
#include <poll.h>

#include "sound.h"

#define	MAX_AUDIO_PORTS	20

struct _sfx {
	long		*soundSize ;
	short		**soundData ;
	int		id ;
	int		loop ;
	int		repeat ;
	int		count ;
	int		pitch ;
	int		enabled ;
	float		loVal ;
	float		hiVal ;
	} ;

typedef struct {
	int		repeat ;
	int		id ;
	long		sampsToPlay ;
	long		sampsPlayed ;
	short		*sample ;
	} Sample ;


/* BEGIN PROTOTYPES -S sound.c */
static int      sfxCheckForAudio( void ) ;
static void     sfxCheckVolume( void ) ;
static void     sfxDieGracefully( int sig, int code, struct sigcontext *sc ) ;
static ALport   sfxInitAudioPort( int init ) ;
static short *  sfxLoadAudioFile( char *dirName, char *fileName, long *size ) ;
static int      sfxOpenAudioPorts( ALport ports[], int maxAudioPorts ) ;
static void     sfxResetAudioHw( void ) ;
static void     sfxSetGain( float l, float r ) ;
static void     sfxSetGainIndex( unsigned long gain ) ;
static void     sfxSetRate( void ) ;
static void     sfxSignalSound( Sfx s ) ;
static void     sfxSoundDied( int sig, int code, struct sigcontext *sc ) ;
static void     sfxSoundErrFunc( long err, const char *msg, ... ) ;
static void     sfxSoundHandler( void *arg ) ;
/* END PROTOTYPES -S sound.c */



static ALport		audioPort[MAX_AUDIO_PORTS] ;
static int		nAudioPorts ;
static int		soundEnabled = 0 ;
static int		soundChild ;
static unsigned long	origLeftGain ;
static unsigned long	origRightGain ;
static unsigned long	origOutputRate ;
static unsigned long	currLeftGain ;
static unsigned long	currRightGain ;
static int		gainIndex ;
static struct sigaction	parentAct ;
static int		endingOnPurpose = 0 ;
static int		spigot[2] ;
static unsigned long	gainSettings[] =
				{ 0, 2, 3, 5, 9, 16, 29, 50, 91, 156, 255 } ;
static char		*sfxAudioDir = "." ;
static int		soundId = 0 ;



/*------------------------------------------------------------------------------
 * Initialize audio.
 *----------------------------------------------------------------------------*/
int
sfxInit(
	char	*audioDir,
	int	numberOfPorts
	)
{
	int	i ;
	int	err ;
	char	*str ;

	soundEnabled = 0 ;

	if( audioDir != NULL )
	{
		str = strdup( audioDir ) ;
		if( str )
		{
			sfxAudioDir = str ;
		}
	}

	if( numberOfPorts < 1 )
	{
		return( SFX_ERR_NO_PORTS_AVAIL ) ;
	}
	else if( numberOfPorts > MAX_AUDIO_PORTS )
	{
		numberOfPorts = MAX_AUDIO_PORTS ;
	}

	(void)ALseterrorhandler( sfxSoundErrFunc ) ;

	nAudioPorts = sfxOpenAudioPorts( audioPort, numberOfPorts ) ;

	if( nAudioPorts == 0 )
	{
		return( SFX_ERR_NO_PORTS_AVAIL ) ;
	}
	else if( nAudioPorts < 0 )
	{
		return( SFX_ERR_NO_AUDIO_HW ) ;
	}

	parentAct.sa_handler = sfxSoundDied ;
	sigemptyset( &(parentAct.sa_mask) ) ;
	parentAct.sa_flags = SA_NOCLDSTOP ;
	sigaction( SIGCHLD, &parentAct, NULL ) ;

	if( pipe( spigot ) < 0 || ( soundChild =
		sproc( sfxSoundHandler, PR_SADDR, audioPort ) ) < 0 )
	{
		for( i = 0 ; i < nAudioPorts ; i++ )
		{
			ALcloseport( audioPort[i] ) ;
		}
		return( SFX_ERR_NO_SPROC ) ;
	}
	else
	{
		soundEnabled = 1 ;
	}

	return( nAudioPorts ) ;
}



/*------------------------------------------------------------------------------
 * Load an audio file.
 *----------------------------------------------------------------------------*/
Sfx
sfxLoad(
	char	*filename
	)
{
	Sfx	s ;

	if( ( s = (Sfx)malloc( sizeof( struct _sfx ) ) ) == NULL )
	{
		return( NULL ) ;
	}
	s->soundData = NULL ;
	s->soundSize = NULL ;
	s->loop = 0 ;
	s->repeat = 0 ;
	s->count = 1 ;
	s->pitch = 0 ;
	s->enabled = 1 ;

	if( ( s->soundData = (short **)malloc( sizeof( short * ) ) ) == NULL )
	{
		sfxFree( s ) ;
		return( NULL ) ;
	}

	if( ( s->soundSize = (long *)malloc( sizeof( long ) ) ) == NULL )
	{
		sfxFree( s ) ;
		return( NULL ) ;
	}

	if( ( s->soundData[0] = sfxLoadAudioFile( sfxAudioDir, filename,
		&(s->soundSize[0]) ) ) == NULL )
	{
		sfxFree( s ) ;
		return( NULL ) ;
	}

	s->id = soundId++ ;

	return( s ) ;
}



/*------------------------------------------------------------------------------
 * Indicate that a sound is to loop.
 *----------------------------------------------------------------------------*/
void
sfxLoop(
	Sfx	s
	)
{
	if( s != NULL )
	{
		s->loop = 1 ;
	}
}



/*------------------------------------------------------------------------------
 * Play sound effect.
 *----------------------------------------------------------------------------*/
void
sfxPlay(
	Sfx	s
	)
{
	/*
	 * If audio is possible.
	 */
	if( soundEnabled && s != NULL && s->soundData[0] && s->enabled )
	{
		if( s->loop )
		{
			s->repeat = 1 ;
		}

		sfxCheckVolume() ;

		sfxSignalSound( s ) ;
	}
}



/*------------------------------------------------------------------------------
 * Play a pitch-bent audio clip.
 *----------------------------------------------------------------------------*/
void
sfxPlayPitch(
	Sfx	s,
	float	val
	)
{
	int	i ;

	if( soundEnabled && s != NULL && s->soundData[0] && s->enabled )
	{
		if( val <= s->loVal )
		{
			i = 0 ;
		}
		else if( val >= s->hiVal )
		{
			i = s->count - 1 ;
		}
		else
		{
			i = (long)( 0.4f + ( s->count - 1 ) *
				( val - s->loVal ) / ( s->hiVal - s->loVal ) ) ;
		}

		if( s->loop && s->repeat && s->pitch == i )
		{
			return ;
		}
		s->pitch = i ;

		sfxPlay( s ) ;
	}
}



/*------------------------------------------------------------------------------
 * Create a pitch bend audio clip.
 *----------------------------------------------------------------------------*/
void
sfxPitchBend(
	Sfx	s,
	float	startVal,
	float	endVal,
	float	startPitch,
	float	endPitch,
	float	basePitch,
	int	steps
	)
{
	int	i ;
	int	size ;
	short	**ptr ;
	short	*data ;
	short	*origData ;
	long	*lptr ;
	double	pitch ;
	double	a ;
	double	b ;
	long	frameIndex ;
	int	k ;

	if( s == NULL || steps < 2 || s->count != 1 )
	{
		return ;
	}

	if( ( lptr = (long *)malloc( steps * sizeof( long ) ) ) == NULL )
	{
		return ;
	}

	size = 0 ;
	a = ( endPitch - startPitch ) / (double)( steps - 1 ) ;
	b = 1.0 / (double)basePitch ;
	for( i = 0 ; i < steps ; i++ )
	{
		pitch = ( startPitch + (double)i * a ) * b ;
		k = (long)( (double)s->soundSize[0] / pitch ) ;
		lptr[i] = k ;
		size += k ;
	}

	if( ( ptr = (short **)malloc( steps * sizeof( short * ) ) ) == NULL )
	{
		free( lptr ) ;
		return ;
	}

	/*
	 * Allocate new space for pitch-bent sound data.
	 */
	if( ( data = (short *)malloc( size * 2 * sizeof( short ) ) ) == NULL )
	{
		free( lptr ) ;
		free( ptr ) ;
		return ;
	}

	size = s->soundSize[0] ;
	free( s->soundSize ) ;
	s->soundSize = lptr ;

	origData = s->soundData[0] ;
	free( s->soundData ) ;
	s->soundData = ptr ;

	s->count = steps ;
	s->loVal = startVal ;
	s->hiVal = endVal ;

	for( i = 0 ; i < steps ; i++ )
	{
		pitch = ( startPitch + (double)i * a ) * b ;

		if( i == 0 )
		{
			s->soundData[i] = data ;
		}
		else
		{
			s->soundData[i] = s->soundData[i-1] +
						2 * s->soundSize[i-1] ;
		}

		for( k = 0 ; k < s->soundSize[i] ; k++ )
		{
			frameIndex = (long)( k * pitch + 0.5 ) ;
			if( frameIndex >= size )
			{
				frameIndex = size - 1 ;
			}

			s->soundData[i][2*k  ] = origData[2*frameIndex  ] ;
			s->soundData[i][2*k+1] = origData[2*frameIndex+1] ;
		}
	}

	free( origData ) ;
}



/*------------------------------------------------------------------------------
 * Silence an audio loop.
 *----------------------------------------------------------------------------*/
void
sfxSilenceLoop(
	Sfx	s
	)
{
	if( s != NULL && s->repeat )
	{
		s->repeat = 0 ;
		sfxSignalSound( s ) ;
	}
}



/*------------------------------------------------------------------------------
 * Disable a sound clip.
 *----------------------------------------------------------------------------*/
void
sfxDisable(
	Sfx	s
	)
{
	if( s != NULL )
	{
		s->enabled = 0 ;
		if( s->loop )
		{
			sfxSilenceLoop( s ) ;
		}
	}
}



/*------------------------------------------------------------------------------
 * Enable a sound clip.
 *----------------------------------------------------------------------------*/
void
sfxEnable(
	Sfx	s
	)
{
	if( s != NULL )
	{
		s->enabled = 1 ;
	}
}



/*------------------------------------------------------------------------------
 * Clean up sound routines.
 *----------------------------------------------------------------------------*/
void
sfxEnd(
	void
	)
{
	endingOnPurpose = 1 ;

	if( soundChild > 0 )
	{
		kill( soundChild, SIGKILL ) ;
	}
	if( nAudioPorts > 0 )
		sfxResetAudioHw() ;
}



/*------------------------------------------------------------------------------
 * Open up the audio ports.
 *----------------------------------------------------------------------------*/
static int
sfxOpenAudioPorts(
	ALport	ports[],
	int	maxAudioPorts
	)
{
	int	n ;

	if( sfxCheckForAudio() == 0 )
	{
		return( -1 ) ;
	}

	for( n = 0 ; n < maxAudioPorts ; n++ )
	{
		if( ( audioPort[n] = sfxInitAudioPort( n == 0 ) ) == NULL )
		{
			break ;
		}
	}

	return( n ) ;
}



/*------------------------------------------------------------------------------
 * Sound handler routine.
 *----------------------------------------------------------------------------*/
static void
sfxSoundHandler(
	void	*arg
	)
{
	ALport			*ap = (ALport *)arg ;
	int			i ;
	int			treated ;
	int			nap ;
	long			sampCount ;
	int			nSounds = 1 ;
	int			nextPort = 0 ;
	long			maxSampsPerPass ;
	Sample			samp[MAX_AUDIO_PORTS] ;
	struct sigaction	act ;
	struct _sfx		ss ;
	struct pollfd		pf ;

	prctl( PR_TERMCHILD, 0 ) ;

	act.sa_handler = sfxDieGracefully ;
	sigemptyset( &(act.sa_mask) ) ;
	act.sa_flags = 0 ;
	sigaction( SIGHUP, &act, NULL ) ;

	maxSampsPerPass = 1600 ;

	for( i = 0 ; i < nAudioPorts ; i++ )
	{
		samp[i].id = -1 ;
		samp[i].sample = NULL ;
		samp[i].sampsToPlay = 0 ;
		samp[i].repeat = 0 ;
	}

	pf.fd = spigot[0] ;
	pf.events = POLLIN | POLLRDNORM | POLLRDBAND ;

	while( 1 )
	{
		if( nSounds == 0 || ( i = poll( &pf, 1, 0 ) ) > 0 )
		{
			read( spigot[0], &ss, sizeof( ss ) ) ;
			treated = 0 ;
			if( ss.loop == 1 && ss.repeat == 0 )
			{
				treated = 1 ;
				for( i = 0 ; i < nAudioPorts ; i++ )
				{
					if( samp[i].id == ss.id )
					{
						samp[i].id = -1 ;
						samp[i].repeat = 0 ;
						samp[i].sampsToPlay = 0 ;
						samp[i].sample = NULL ;
					}
				}
			}
			else if( ss.loop == 1 && ss.count > 1 )
			{
				for( i = 0 ; i < nAudioPorts ; i++ )
				{
					if( samp[i].id == ss.id )
					{
						treated = 1 ;
						samp[i].repeat = ss.repeat ;
						samp[i].sampsToPlay =
							ss.soundSize[ss.pitch] ;
						samp[i].sampsPlayed = 0 ;
						samp[i].sample =
							ss.soundData[ss.pitch] ;
					}
				}
			}
			if( !treated )
			{
				for( i = 0 ; i < nAudioPorts ; i++ )
				{
					nextPort = ( nextPort + 1 ) %
							nAudioPorts ;
					if( samp[nextPort].repeat == 0 )
					{
						break ;
					}
				}
				samp[nextPort].sample = ss.soundData[ss.pitch] ;
				samp[nextPort].sampsToPlay =
							ss.soundSize[ss.pitch] ;
				samp[nextPort].sampsPlayed = 0 ;
				samp[nextPort].repeat = ss.repeat ;
				samp[nextPort].id = ss.id ;
			}
		}
		else if ( i < 0 )
		{
			perror( "input poll" ) ;
		}
		nSounds = 0 ;
		nap = 0 ;
		sfxSetRate() ;
		for( i = 0 ; i < nAudioPorts ; i++ )
		{
			if( samp[i].sampsToPlay > 0 )
			{
				nSounds++ ;
				if( ALgetfilled( ap[i] ) > 4000 )
				{
					nap++ ;
					continue ;
				}
			}
			if( samp[i].sampsToPlay > maxSampsPerPass )
			{
				if( ALwritesamps( ap[i],
					samp[i].sample + samp[i].sampsPlayed,
					maxSampsPerPass ) < 0 )
				{
					printf( "ALwritesamps error %d\n",
						oserror() ) ;
					exit( 1 ) ;
				}
				samp[i].sampsPlayed += maxSampsPerPass ;
				samp[i].sampsToPlay -= maxSampsPerPass ;
			}
			else if( samp[i].sampsToPlay > 0 )
			{
				ALwritesamps( ap[i],
					samp[i].sample + samp[i].sampsPlayed,
					samp[i].sampsToPlay ) ;
				if( samp[i].repeat )
				{
					sampCount = maxSampsPerPass -
							samp[i].sampsToPlay ;
					samp[i].sampsToPlay +=
						samp[i].sampsPlayed - sampCount;
					samp[i].sampsPlayed = sampCount ;
					ALwritesamps( ap[i], samp[i].sample,
						sampCount ) ;
				}
				else
				{
					samp[i].sampsToPlay = 0 ;
				}
			}
		}
		if( nap == nSounds )
		{
			sginap( 1 ) ;
		}
	}
}



/*------------------------------------------------------------------------------
 * Error handler function for SIGHUP
 *----------------------------------------------------------------------------*/
static void
sfxDieGracefully(
	int			sig,
	int			code,
	struct sigcontext	*sc
	)
{
	exit( 0 ) ;
}



/*------------------------------------------------------------------------------
 * Reset the audio hardware to where we found it when we started.
 *----------------------------------------------------------------------------*/
static void
sfxResetAudioHw(
	void
	)
{
	long	pvbuf[6] ;
	long	buflen ;

	pvbuf[0] = AL_LEFT_SPEAKER_GAIN ;
	pvbuf[1] = origLeftGain ;
	pvbuf[2] = AL_RIGHT_SPEAKER_GAIN ;
	pvbuf[3] = origRightGain ;
	pvbuf[4] = AL_OUTPUT_RATE ;
	pvbuf[5] = origOutputRate ;
	buflen = 6 ;

	ALsetparams( AL_DEFAULT_DEVICE, pvbuf, buflen ) ;
}



/*------------------------------------------------------------------------------
 * Open and initialize an audio port.
 *----------------------------------------------------------------------------*/
static ALport
sfxInitAudioPort(
	int	init
	)
{
	long		pvbuf[6] ;
	long		buflen ;
	ALconfig	audioPortConfig;
	ALport		ap ;

	if( init )
	{
		pvbuf[0] = AL_LEFT_SPEAKER_GAIN ;
		pvbuf[2] = AL_RIGHT_SPEAKER_GAIN ;
		pvbuf[4] = AL_OUTPUT_RATE ;
		buflen = 6 ;
		ALgetparams( AL_DEFAULT_DEVICE, pvbuf, buflen ) ;
		currLeftGain = origLeftGain = pvbuf[1] ;
		currRightGain = origRightGain = pvbuf[3] ;
		origOutputRate = pvbuf[5] ;

		sfxSetGainIndex( ( origLeftGain + origRightGain ) / 2 ) ;
	}

	/*
	 * configure and open audio port
	 */
	audioPortConfig = ALnewconfig();
	ALsetwidth( audioPortConfig, AL_SAMPLE_16 ) ;
	ALsetchannels( audioPortConfig, AL_STEREO ) ;

	ALsetqueuesize( audioPortConfig, 10000 ) ;
	ap = ALopenport( "sfx", "w", audioPortConfig ) ;

	ALfreeconfig( audioPortConfig ) ;

	return( ap ) ;
}



/*------------------------------------------------------------------------------
 * Set the gain index.
 *----------------------------------------------------------------------------*/
static void
sfxSetGainIndex(
	unsigned long	gain
	)
{
	int	i ;
	long	min = 256 ;
	long	diff ;

	for( i = 0 ; i < sizeof( gainSettings ) / sizeof( gainSettings[0] ) ;
		i++ )
	{
		diff = gain - gainSettings[i] ;
		if( diff < 0 )
		{
			diff = -diff ;
		}
		if( diff < min )
		{
			min = diff ;
			gainIndex = i ;
		}
	}
}



/*------------------------------------------------------------------------------
 * Set the volume level.
 *----------------------------------------------------------------------------*/
static void
sfxSetGain(
	float l,
	float r
	)
{
	long	pvbuf[6] ;
	long	buflen ;

	pvbuf[0] = AL_LEFT_SPEAKER_GAIN ;
	pvbuf[1] = l * currLeftGain ;
	pvbuf[2] = AL_RIGHT_SPEAKER_GAIN ;
	pvbuf[3] = r * currRightGain ;
	pvbuf[4] = AL_OUTPUT_RATE ;
	pvbuf[5] = AL_RATE_16000 ;
	buflen = 6 ;

	ALsetparams( AL_DEFAULT_DEVICE, pvbuf, buflen ) ;
}



/*------------------------------------------------------------------------------
 * Set the audio rate.
 *----------------------------------------------------------------------------*/
static void
sfxSetRate(
	void
	)
{
	long	pvbuf[2] ;
	long	buflen ;

	pvbuf[0] = AL_OUTPUT_RATE ;
	pvbuf[1] = AL_RATE_16000 ;
	buflen = 2 ;

	ALsetparams( AL_DEFAULT_DEVICE, pvbuf, buflen ) ;
}



/*------------------------------------------------------------------------------
 * Check for audio hardware.
 *----------------------------------------------------------------------------*/
static int
sfxCheckForAudio(
	void
	)
{
	int		st = 0 ;
	inventory_t	*base ;
	inventory_t	*inv ;

	inv = base = getinvent() ;
	while( inv != NULL )
	{
		/*
		 * Ok if any type of audio hardware available.
		 */
#if defined( __SVR3 )
		if( inv->class == INV_AUDIO )
#else
		if( inv->inv_class == INV_AUDIO )
#endif /* defined( __SVR3 ) */
		{
			st = 1 ;
		}
		inv = getinvent() ;
	}
	
	if( base )
		endinvent() ;

	return( st ) ;
}



/*------------------------------------------------------------------------------
 * Open an audio file, check for recognized type, and read.
 *----------------------------------------------------------------------------*/
static short *
sfxLoadAudioFile(
	char	*dirName,
	char	*fileName,
	long	*size
	)
{
	int		fd ;
	AFfilehandle	file ;
	int		framesPerSec ;
	long		bitsPerSamp ;
	long		sampType ;
	char		*name ;
	short		*data ;

	name = malloc( strlen( dirName ) + strlen( fileName ) + 2 ) ;
	sprintf( name, "%s/%s", dirName, fileName ) ;

	fd = open( name, O_RDONLY ) ;

	if( fd < 0 )
	{
		perror( name ) ;
		free( name ) ;
		return( NULL ) ;
	}

	switch( AFidentifyfd( fd ) )
	{
		case AF_FILE_AIFF :
		case AF_FILE_AIFFC :
			break ;

		default :
			fprintf( stderr, "%s: unrecognized file type -- convert"
				" to AIFF or AIFC\n", name ) ;
			free( name ) ;
			return( NULL ) ;
	}

	file = AFopenfd( fd, "r", AF_NULL_FILESETUP ) ;

	if( file == AF_NULL_FILEHANDLE )
	{
		fprintf( stderr, "%s: failed to attach an audio file struct\n",
			 name ) ;
		free( name ) ;
		return( NULL ) ;
	}

	if( (int)AFgetchannels( file, AF_DEFAULT_TRACK ) != 2 )
	{
		fprintf( stderr, "%s: does not have 2 channels\n", name ) ;
		free( name ) ;
		AFclosefile( file ) ;
		return( NULL ) ;
		}

	if( ( framesPerSec = AFgetrate( file, AF_DEFAULT_TRACK ) ) != 16000 )
	{
		fprintf( stderr, "%s: is not recorded at 16 KHz\n", name ) ;
		free( name ) ;
		AFclosefile( file ) ;
		return( NULL ) ;
	}

	AFgetsampfmt( file, AF_DEFAULT_TRACK, &sampType, &bitsPerSamp ) ;
	if( bitsPerSamp != 16 )
	{
		fprintf( stderr, "%s: is not recorded with 16 bit samples\n",
			name ) ;
		free( name ) ;
		AFclosefile( file ) ;
		return( NULL ) ;
	}

	free( name ) ;

	*size = AFgetframecnt( file, AF_DEFAULT_TRACK ) ;

	if( ( data = (short *)malloc( *size * 2 * sizeof( short ) ) ) == NULL )
	{
		fprintf( stderr, "%s: out of memory for %d samples\n",
			name, *size ) ;
		*size = 0 ;
		AFclosefile( file ) ;
		return( NULL ) ;
	}

	if( AFreadframes( file, AF_DEFAULT_TRACK, data, *size ) != *size )
	{
		fprintf( stderr, "%s: error reading\n", name ) ;
		free( data ) ;
		data = NULL ;
		*size = 0 ;
		}

	*size *= 2 ;

	AFclosefile( file ) ;

	return( data ) ;
}



/*------------------------------------------------------------------------------
 * Wake up the sound handler.
 *----------------------------------------------------------------------------*/
static void
sfxSignalSound(
	Sfx	s
	)
{
	struct _sfx	ss ;

	if( soundChild )
	{
		if( write( spigot[1], s, sizeof( *s ) ) != sizeof( *s ) )
		{
			perror( "sfxSignalSound" ) ;
		}
	}
}



/*------------------------------------------------------------------------------
 * Adjust the volume.
 *----------------------------------------------------------------------------*/
int
sfxVolumeChange(
	int	direction
	)
{
	if( soundEnabled )
	{
		sfxCheckVolume() ;
		if( direction == 1 )
		{
			if( gainIndex < sizeof( gainSettings ) /
				sizeof( gainSettings[0] ) - 1 )
			{
				gainIndex++ ;
			}
		}
		else
		{
			if( gainIndex > 0 )
			{
				gainIndex-- ;
			}
		}
		currLeftGain = currRightGain = gainSettings[gainIndex] ;

		sfxSetGain( 1., 1. ) ;

		return( gainIndex ) ;
	}
	else
	{
		return( 0 ) ;
	}
}



/*------------------------------------------------------------------------------
 * Handler for sound child dying.
 *----------------------------------------------------------------------------*/
static void
sfxSoundDied(
	int			sig,
	int			code,
	struct sigcontext	*sc
	)
{
	if( !endingOnPurpose )
	{
		nAudioPorts = 0 ;
		soundEnabled = 0 ;

		fprintf( stderr, "Sound handling child process was killed by"
			" signal %d (code = %d)\n", sig, code ) ;
	}
}



/*------------------------------------------------------------------------------
 * Set the volume to a specified level.
 *----------------------------------------------------------------------------*/
void
sfxSetVolume(
	int	level
	)
{
	if( level >= 0 &&
		level < sizeof( gainSettings ) / sizeof( gainSettings[0] ) )
	{
		gainIndex = level ;
		currLeftGain = currRightGain = gainSettings[gainIndex] ;
	}
}



/*------------------------------------------------------------------------------
 * Free the Sfx structure.
 *----------------------------------------------------------------------------*/
void
sfxFree(
	Sfx	s
	)
{
	if( s )
	{
		if( s->soundData )
		{
			free( s->soundData ) ;
			s->soundData = NULL ;
		}

		if( s->soundSize )
		{
			free( s->soundSize ) ;
			s->soundSize = NULL ;
		}

		free( s ) ;
	}
}



/*------------------------------------------------------------------------------
 * Audio error function handler.
 *----------------------------------------------------------------------------*/
static void
sfxSoundErrFunc(
	long		err,
	const char	*msg,
	...
	)
{
}



/*------------------------------------------------------------------------------
 * Check for external volume changes.
 *----------------------------------------------------------------------------*/
static void
sfxCheckVolume(
	void
	)
{
	long	pvbuf[6] ;
	long	buflen ;

	/*
	 * Check to see if the volume was changed externally.
	 */
	pvbuf[0] = AL_LEFT_SPEAKER_GAIN ;
	pvbuf[2] = AL_RIGHT_SPEAKER_GAIN ;
	buflen = 4 ;

	ALgetparams( AL_DEFAULT_DEVICE, pvbuf, buflen ) ;
	if( pvbuf[1] != currLeftGain || pvbuf[3] != currRightGain )
	{
		origLeftGain = currLeftGain = pvbuf[1] ;
		origRightGain = currRightGain = pvbuf[3] ;
		sfxSetGainIndex( ( pvbuf[1] + pvbuf[3] ) / 2 ) ;
	}
}
