/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
#include <stdio.h>
#include <math.h>
#include <malloc.h>

#include "vroom.h"
#include "track.h"
#include "car.h"
#include "body.h"


#define	DXDS(t)		( (t)->c[0] )
#define	DYDS(t)		( (t)->c[1] )
#define	DXDL(t)		( (t)->c[2] )
#define	DYDL(t)		( (t)->c[3] )

/* BEGIN PROTOTYPES -S straight.c */
static void     addedStartingTrack( Track *track ) ;
static void     addedStraightTrack( Track *track ) ;
static void     addToStraightTrack( Track *old, Track *new ) ;
static void     drawSideMarkers( GLfloat w, GLfloat c, GLfloat s,
                    Track *track ) ;
static void     drawStartingTrack( Track *track, int lod ) ;
static void     drawStartingTrackMarkers( Track *track, int lod ) ;
static void     drawStartOutline( Track *track, int side ) ;
static void     drawStraightOutline( Track *track, int side ) ;
static void     drawStraightTrack( Track *track, int lod ) ;
static void     drawStraightTrackMarkers( Track *track, int lod ) ;
static void     endPosStraightTrack( Track *track, Car *car, int place ) ;
static void     followStraightTrack( Track *track, Car *car, float dTime ) ;
static void     freeStraightTrack( Track *track ) ;
static float    getStraightLength( Track *track, float lane ) ;
static void     startPosStraightTrack( Track *track, Car *car, int place ) ;
static float    straightLaneWidth( Track *track, float s, float lane1,
                    float lane2 ) ;
static float    topSpeedStraightTrack( Track *track, float s, float l ) ;
/* END PROTOTYPES -S straight.c */


extern int	debugOn ;
extern float	carLength ;



/*------------------------------------------------------------------------------
 * Follow straight track.
 *----------------------------------------------------------------------------*/
static void
followStraightTrack(
	Track	*track,
	Car	*car,
	float	dTime
	)
{
	int	doNext = 0 ;
	int	wreckDir = 0 ;
	float	eTime ;
	float	newS ;
	float	partialLane ;
	float	wreckVel[2] ;

	if( car->lane < 1.0f || car->lane > (float)track->lanes )
	{
		if( car->lane < 1.0f )
		{
			wreckVel[0] = cosf( track->theta + M_PI * 0.6667f );
			wreckVel[1] = sinf( track->theta + M_PI * 0.6667f );
			wreckDir = -1 ;
		}
		else
		{
			wreckVel[0] = cosf( track->theta + M_PI * 0.3333f ) ;
			wreckVel[1] = sinf( track->theta + M_PI * 0.3333f ) ;
			wreckDir = 1 ;
		}
		wipeOutCar( car, wreckVel[0], wreckVel[1], wreckDir,
				WIPE_OUT_OFF_ROAD ) ;
	}
	else
	{
		if( car->status & CAR_SKID )
		{
			stopSkid( car ) ;
		}
		newS = car->s + car->speed * dTime / track->length[0] ;
		if( newS >= 1.0f )
		{
			eTime = ( 1.0f - car->s ) * track->length[0] /
				car->speed ;
			dTime -= eTime ;
			car->s = 0.0f ;
			car->track = track->nextTrack ;
			incrementLap( car, dTime ) ;
			car->track->enter( car->track, car ) ;
			doNext = 1 ;
		}
		else
		{
			eTime = dTime ;
			car->s = newS ;
		}
		if( car->status & CHANGE_LANE_RIGHT )
		{
			partialLane = car->lane - floorf( car->lane ) ;
			partialLane += eTime * LANE_CHANGE_SPEED ;
			if( partialLane > 1.0f )
			{
				car->status &= ~CHANGE_LANE_RIGHT ;
				car->lane = floorf( car->lane ) + 1.0f ;
			}
			else
			{
				car->lane = floorf( car->lane ) + partialLane ;
			}
		}
		else if( car->status & CHANGE_LANE_LEFT )
		{
			partialLane = ceilf( car->lane ) - car->lane ;
			partialLane += eTime * LANE_CHANGE_SPEED ;
			if( partialLane > 1.0f )
			{
				car->status &= ~CHANGE_LANE_LEFT ;
				car->lane = ceilf( car->lane ) - 1.0f ;
			}
			else
			{
				car->lane = ceilf( car->lane ) - partialLane ;
			}
		}
		if( doNext )
		{
			car->track->follow( car->track, car, dTime ) ;
		}
		else
		{
			car->x = track->x + car->s * DXDS( track ) +
				( car->lane - track->midLane ) *
				DXDL( track ) ;
			car->y = track->y + car->s * DYDS( track ) +
				( car->lane - track->midLane ) *
				DYDL( track ) ;
			car->theta = track->theta ;
		}
	}
}



/*------------------------------------------------------------------------------
 * Initialize after adding track.
 *----------------------------------------------------------------------------*/
static void
addedStraightTrack(
	Track	*track
	)
{
	int	i ;
	float	radius ;

	DXDS( track ) = -sinf( track->theta ) * track->length[0] ;
	DYDS( track ) = cosf( track->theta ) * track->length[0] ;
	DXDL( track ) = cosf( track->theta ) * LANE_WIDTH ;
	DYDL( track ) = sinf( track->theta ) * LANE_WIDTH ;

	track->midLane = 1.0f + 0.5f * ( track->lanes - 1.0f ) ;

	for( i = 0 ; i < MAX_LOD ; i++ )
	{
		glNewList( track->drawList+i, GL_COMPILE ) ;
		drawStraightTrack( track, i ) ;
		glEndList() ;
		glNewList( track->drawMarkerList+i, GL_COMPILE ) ;
		drawStraightTrackMarkers( track, i ) ;
		glEndList() ;
	}

	track->center[0] = track->x + 0.5f * DXDS( track ) ;
	track->center[1] = track->y + 0.5f * DYDS( track ) ;
	track->center[2] = 0.0f ;

	radius = 0.5f * track->lanes * LANE_WIDTH + SIDE_MARKER_WIDTH ;
	radius = 0.5f * sqrtf( track->length[0] * track->length[0] +
		radius * radius ) ;
	track->cullDetail = trackDetail( radius ) ;
}



/*------------------------------------------------------------------------------
 * Create a N lane piece of straight track.
 *----------------------------------------------------------------------------*/
Track *
createStraightTrack(
	int	nLanes,
	float	length
	)
{
	int	i ;
	Track	*track ;

	track = myMalloc( sizeof( Track ) ) ;

	track->lanes = nLanes ;
	track->type = STRAIGHT_TRACK ;
	track->x = 0.0f ;
	track->y = 0.0f ;
	track->theta = 0.0f ;
	track->length = myMalloc( nLanes * sizeof( float ) ) ;
	for( i = 0 ; i < nLanes ; i++ )
	{
		track->length[i] = length ;
	}
	track->follow = followStraightTrack ;
	track->add = addToStraightTrack ;
	track->added = addedStraightTrack ;
	track->topSpeed = topSpeedStraightTrack ;
	track->nextTrack = NULL ;
	track->drawList = glGenLists( MAX_LOD ) ;
	track->drawMarkerList = glGenLists( MAX_LOD ) ;
	track->startPosition = startPosStraightTrack ;
	track->endPosition = endPosStraightTrack ;
	track->drawOutline = drawStraightOutline ;
	track->trackLength = getStraightLength ;
	track->enter = enterTrack ;
	track->free = freeStraightTrack ;
	track->cullDetail = NULL ;
	track->laneWidth = straightLaneWidth ;

	return( track ) ;
}



/*------------------------------------------------------------------------------
 * Create a N lane piece of starting/finishing track.
 *----------------------------------------------------------------------------*/
Track *
createStartingTrack(
	int	nLanes,
	float	length
	)
{
	Track	*track ;

	track = createStraightTrack( nLanes, length ) ;
	track->added = addedStartingTrack ;
	track->drawOutline = drawStartOutline ;

	return( track ) ;
}



/*------------------------------------------------------------------------------
 * Initialize after adding track.
 *----------------------------------------------------------------------------*/
static void
addedStartingTrack(
	Track	*track
	)
{
	int	i ;
	float	radius ;

	DXDS( track ) = -sinf( track->theta ) * track->length[0] ;
	DYDS( track ) = cosf( track->theta ) * track->length[0] ;
	DXDL( track ) = cosf( track->theta ) * LANE_WIDTH ;
	DYDL( track ) = sinf( track->theta ) * LANE_WIDTH ;

	track->midLane = 1.0f + 0.5f * ( track->lanes - 1.0f ) ;

	for( i = 0 ; i < MAX_LOD ; i++ )
	{
		glNewList( track->drawList+i, GL_COMPILE ) ;
		drawStartingTrack( track, i ) ;
		glEndList() ;
		glNewList( track->drawMarkerList+i, GL_COMPILE ) ;
		drawStartingTrackMarkers( track, i ) ;
		glEndList() ;
	}

	track->center[0] = track->x + 0.5f * DXDS( track ) ;
	track->center[1] = track->y + 0.5f * DYDS( track ) ;
	track->center[2] = 0.0f ;

	radius = 0.5f * track->lanes * LANE_WIDTH + SIDE_MARKER_WIDTH ;
	radius = 0.5f * sqrtf( track->length[0] * track->length[0] +
		radius * radius ) ;
	track->cullDetail = trackDetail( radius ) ;
}



/*------------------------------------------------------------------------------
 * Add to a piece of straight track.
 *----------------------------------------------------------------------------*/
static void
addToStraightTrack(
	Track	*old,
	Track	*new
	)
{
	new->theta = old->theta ;
	new->x = old->x + DXDS( old ) ;
	new->y = old->y + DYDS( old ) ;
	old->nextTrack = new ;
}



/*------------------------------------------------------------------------------
 * Draw the track piece.
 *----------------------------------------------------------------------------*/
static void
drawStraightTrack(
	Track	*track,
	int	lod
	)
{
	int	i ;
	int	l ;
	GLfloat	c ;
	GLfloat	s ;
	GLfloat	w ;
	GLfloat	a[2] ;
	GLfloat	v[4][2] ;

	c = cosf( track->theta ) ;
	s = sinf( track->theta ) ;

	v[0][1] = 0.0f ;
	v[0][1] = 0.0f ;
	v[2][1] = track->length[0] ;
	v[3][1] = track->length[0] ;

	for( l = 0 ; l < track->lanes ; l++ )
	{
		w = ( (float)l - 0.5f * ( track->lanes - 1 ) ) * LANE_WIDTH ;
		v[0][0] = w - 0.5f * RAIL_WIDTH ;
		v[1][0] = w + 0.5f * RAIL_WIDTH ;
		v[2][0] = w - 0.5f * RAIL_WIDTH ;
		v[3][0] = w + 0.5f * RAIL_WIDTH ;
		glBegin( GL_TRIANGLE_STRIP ) ;
		for( i = 0 ; i < 4 ; i++ )
		{
			glVertex2f( track->x + v[i][0] * c - v[i][1] * s,
				    track->y + v[i][0] * s + v[i][1] * c ) ;
		}
		glEnd() ;
	}
}



/*------------------------------------------------------------------------------
 * Draw the track markers.
 *----------------------------------------------------------------------------*/
static void
drawStraightTrackMarkers(
	Track	*track,
	int	lod
	)
{
	if( lod < MAX_LOD - 1 )
	{
		drawSideMarkers( 0.5f * track->lanes * LANE_WIDTH,
				cosf( track->theta ), sinf( track->theta ),
				track ) ;
	}
}



/*------------------------------------------------------------------------------
 * Draw the straight track piece outline.
 *----------------------------------------------------------------------------*/
static void
drawStraightOutline(
	Track	*track,
	int	side
	)
{
	GLfloat	c ;
	GLfloat	s ;
	GLfloat	w ;
	GLfloat	a[2] ;
	GLfloat	v[2] ;

	w = 0.5f * track->lanes * LANE_WIDTH ;
	v[0] = ( side == TRACK_LEFT_SIDE ) ? -w : w ; v[1] = track->length[0] ;

	c = cosf( track->theta ) ;
	s = sinf( track->theta ) ;
	a[0] = track->x + v[0] * c - v[1] * s ;
	a[1] = track->y + v[0] * s + v[1] * c ;

	glVertex2fv( a ) ;
}



/*------------------------------------------------------------------------------
 * Draw the starting track piece outline.
 *----------------------------------------------------------------------------*/
static void
drawStartOutline(
	Track	*track,
	int	side
	)
{
	int	i ;
	GLfloat	c ;
	GLfloat	s ;
	GLfloat	w ;
	GLfloat	a[2] ;
	GLfloat	v[7][2] ;
	GLfloat	dl = 0.2f * LANE_WIDTH ;

	w = ( ( side == TRACK_LEFT_SIDE ) ? -0.5f : 0.5f ) * track->lanes *
		LANE_WIDTH ;
	v[0][0] = w    ; v[0][1] = 0.0f ;
	v[1][0] = w    ; v[1][1] = track->length[0] - 2.0f * dl ;
	v[2][0] = 0.0f ; v[2][1] = track->length[0] ;
	v[3][0] = w    ; v[3][1] = track->length[0] ;
	v[4][0] = 0.0f ; v[4][1] = v[1][1] ;
	v[5][0] = w    ; v[5][1] = v[1][1] ;
	v[6][0] = w    ; v[6][1] = track->length[0] ;

	c = cosf( track->theta ) ;
	s = sinf( track->theta ) ;
	for( i = 0 ; i < 7 ; i++ )
	{
		a[0] = track->x + v[i][0] * c - v[i][1] * s ;
		a[1] = track->y + v[i][0] * s + v[i][1] * c ;
		glVertex2fv( a ) ;
	}
}



/*------------------------------------------------------------------------------
 * Draw the starting track piece.
 *----------------------------------------------------------------------------*/
static void
drawStartingTrack(
	Track	*track,
	int	lod
	)
{
	int	i ;
	int	l ;
	GLfloat	c ;
	GLfloat	s ;
	GLfloat	w ;
	GLfloat	v[4][2] ;

	c = cosf( track->theta ) ;
	s = sinf( track->theta ) ;

	v[0][1] = 0.0f ;
	v[1][1] = 0.0f ;
	v[2][1] = track->length[0] ;
	v[3][1] = track->length[0] ;

	for( l = 0 ; l < track->lanes ; l++ )
	{
		w = ( (float)l - 0.5f * ( track->lanes - 1 ) ) * LANE_WIDTH ;
		v[0][0] = w - 0.5f * RAIL_WIDTH ;
		v[1][0] = w + 0.5f * RAIL_WIDTH ;
		v[2][0] = w - 0.5f * RAIL_WIDTH ;
		v[3][0] = w + 0.5f * RAIL_WIDTH ;
		glBegin( GL_TRIANGLE_STRIP ) ;
		for( i = 0 ; i < 4 ; i++ )
		{
			glVertex2f( track->x + v[i][0] * c - v[i][1] * s,
				    track->y + v[i][0] * s + v[i][1] * c ) ;
		}
		glEnd() ;
	}
}



/*------------------------------------------------------------------------------
 * Draw the starting track markers.
 *----------------------------------------------------------------------------*/
static void
drawStartingTrackMarkers(
	Track	*track,
	int	lod
	)
{
	GLfloat	dl = 0.2f * LANE_WIDTH ;
	GLfloat	c ;
	GLfloat	s ;
	GLfloat	w ;
	GLfloat	a[2] ;

	w = 0.5f * track->lanes * LANE_WIDTH ;
	c = cosf( track->theta ) ;
	s = sinf( track->theta ) ;

	if( lod < MAX_LOD - 1 )
	{
		drawSideMarkers( w, c, s, track ) ;
	}

	glBegin( GL_TRIANGLE_STRIP ) ;
	glColor3f( 0.9f, 0.9f, 0.9f ) ;
	a[0] = -w + dl ;
	a[1] = track->length[0] - 3.0f * dl ;
	glVertex2f( track->x + a[0] * c - a[1] * s,
		    track->y + a[0] * s + a[1] * c ) ;
	a[0] = -w ;
	a[1] = track->length[0] - 4.0f * dl ;
	glVertex2f( track->x + a[0] * c - a[1] * s,
		    track->y + a[0] * s + a[1] * c ) ;
	a[0] = w - dl ;
	a[1] = track->length[0] - 3.0f * dl ;
	glVertex2f( track->x + a[0] * c - a[1] * s,
		    track->y + a[0] * s + a[1] * c ) ;
	a[0] = w ;
	a[1] = track->length[0] - 4.0f * dl ;
	glVertex2f( track->x + a[0] * c - a[1] * s,
		    track->y + a[0] * s + a[1] * c ) ;
	a[0] = w - dl ;
	a[1] = track->length[0] - 1.0f * dl ;
	glVertex2f( track->x + a[0] * c - a[1] * s,
		    track->y + a[0] * s + a[1] * c ) ;
	a[0] = w ;
	a[1] = track->length[0] ;
	glVertex2f( track->x + a[0] * c - a[1] * s,
		    track->y + a[0] * s + a[1] * c ) ;
	a[0] = -w + dl ;
	a[1] = track->length[0] - 1.0f * dl ;
	glVertex2f( track->x + a[0] * c - a[1] * s,
		    track->y + a[0] * s + a[1] * c ) ;
	a[0] = -w ;
	a[1] = track->length[0] ;
	glVertex2f( track->x + a[0] * c - a[1] * s,
		    track->y + a[0] * s + a[1] * c ) ;
	a[0] = -w + dl ;
	a[1] = track->length[0] - 3.0f * dl ;
	glVertex2f( track->x + a[0] * c - a[1] * s,
		    track->y + a[0] * s + a[1] * c ) ;
	a[0] = -w ;
	a[1] = track->length[0] - 4.0f * dl ;
	glVertex2f( track->x + a[0] * c - a[1] * s,
		    track->y + a[0] * s + a[1] * c ) ;
	glEnd() ;
}



/*------------------------------------------------------------------------------
 * Return the top speed of a straight track.
 *----------------------------------------------------------------------------*/
static float
topSpeedStraightTrack(
	Track	*track,
	float	s,
	float	l
	)
{
	return( 1.5f * TOP_SPEED ) ;
}



/*------------------------------------------------------------------------------
 * Determine starting position of cars.
 *----------------------------------------------------------------------------*/
static void
startPosStraightTrack(
	Track	*track,
	Car	*car,
	int	place
	)
{
	int	row ;

	car->lane = ( place % track->lanes ) + 1 ;

	row = place / track->lanes ;

	car->s = 1.0f - ( (float)row * 1.1f + 0.5f ) * carLength /
			track->length[0] ;

	track->follow( track, car, 0.0f ) ;
}



/*------------------------------------------------------------------------------
 * Determine ending position of cars.
 *----------------------------------------------------------------------------*/
static void
endPosStraightTrack(
	Track	*track,
	Car	*car,
	int	place
	)
{
	int	row ;

	car->lane = 0 ;

	row = place - 1 ;

	car->s = 1.0f - ( (float)row * 1.1f + 0.5f ) * carLength /
			track->length[0] ;

	car->x = track->x + car->s * DXDS( track ) +
		( car->lane - track->midLane ) *
		DXDL( track ) ;
	car->y = track->y + car->s * DYDS( track ) +
		( car->lane - track->midLane ) *
		DYDL( track ) ;
	car->theta = track->theta ;
}



/*------------------------------------------------------------------------------
 * Free up memory used by a piece of track.
 *----------------------------------------------------------------------------*/
static void
freeStraightTrack(
	Track	*track
	)
{
	if( track )
	{
		free( track->length ) ;
		free( track ) ;
	}
}



/*------------------------------------------------------------------------------
 * Return the length of the track.
 *----------------------------------------------------------------------------*/
static float
getStraightLength(
	Track	*track,
	float	lane
	)
{
	return( track->length[0] ) ;
}



/*------------------------------------------------------------------------------
 * Draw the side markers on the track.
 *----------------------------------------------------------------------------*/
static void
drawSideMarkers(
	GLfloat	w,
	GLfloat	c,
	GLfloat	s,
	Track	*track
	)
{
	int	i ;
	int	k ;
	int	n ;
	GLfloat	h ;
	GLfloat	a[2] ;
	GLfloat	v[2][2] ;

	n = (int)( track->length[0] / SIDE_MARKER_LENGTH ) ;
	n += ( 4 - ( n % 4 ) ) % 4 ;
	h = track->length[0] / n ;
	v[0][0] = -w - SIDE_MARKER_WIDTH ;
	v[1][0] = -w ;
	for( k = 0 ; k < 2 ; k++ )
	{
		glBegin( GL_QUADS ) ;
		for( i = 0 ; i < n ; i += 2 )
		{
			if( ( i % 4 ) == 0 )
			{
				glColor3f( 0.5f, 0.0f, 0.0f ) ;
			}
			else
			{
				glColor3f( 0.9f, 0.9f, 0.9f ) ;
			}
			v[0][1] = v[1][1] = i * h ;
			glVertex2f( track->x + v[0][0] * c - v[0][1] * s,
				    track->y + v[0][0] * s + v[0][1] * c ) ;
			glVertex2f( track->x + v[1][0] * c - v[1][1] * s,
				    track->y + v[1][0] * s + v[1][1] * c ) ;
			v[0][1] = v[1][1] += h ;
			glVertex2f( track->x + v[1][0] * c - v[1][1] * s,
				    track->y + v[1][0] * s + v[1][1] * c ) ;
			glVertex2f( track->x + v[0][0] * c - v[0][1] * s,
				    track->y + v[0][0] * s + v[0][1] * c ) ;
		}
		glEnd() ;
		v[0][0] = w ;
		v[1][0] = w + SIDE_MARKER_WIDTH ;
	}
}



/*------------------------------------------------------------------------------
 * Return the distance to a lane.
 *----------------------------------------------------------------------------*/
static float
straightLaneWidth(
	Track	*track,
	float	s,
	float	lane1,
	float	lane2
	)
{
	if( lane1 < 1.0f )
	{
		lane1 = 1.0f ;
	}
	else if( lane1 > (float)track->lanes )
	{
		lane1 = (float)track->lanes ;
	}
	if( lane2 < 1.0f )
	{
		lane2 = 1.0f ;
	}
	else if( lane2 > (float)track->lanes )
	{
		lane2 = (float)track->lanes ;
	}

	lane1 = ( lane1 - lane2 ) * LANE_WIDTH ;
	return( ABSFUNC( lane1 ) ) ;
}
